#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Implementation of a base container registry class."""

import re
from typing import TYPE_CHECKING, Optional, Tuple, Type, cast

from pydantic import Field, field_validator

from zenml.constants import DOCKER_REGISTRY_RESOURCE_TYPE
from zenml.enums import StackComponentType
from zenml.models import ServiceConnectorRequirements
from zenml.secret.schemas import BasicAuthSecretSchema
from zenml.stack.authentication_mixin import (
    AuthenticationConfigMixin,
    AuthenticationMixin,
)
from zenml.stack.flavor import Flavor
from zenml.utils import docker_utils

if TYPE_CHECKING:
    from docker.client import DockerClient


class BaseContainerRegistryConfig(AuthenticationConfigMixin):
    """Base config for a container registry.

    Configuration for connecting to container image registries.
    Field descriptions are defined inline using Field() descriptors.
    """

    uri: str = Field(
        description="Container registry URI (e.g., 'gcr.io' for Google Container "
        "Registry, 'docker.io' for Docker Hub, 'registry.gitlab.com' for GitLab "
        "Container Registry, 'ghcr.io' for GitHub Container Registry). This is "
        "the base URL where container images will be pushed to and pulled from."
    )
    default_repository: Optional[str] = Field(
        default=None,
        description="Default repository namespace for image storage (e.g., "
        "'username' for Docker Hub, 'project-id' for GCR, 'organization' for "
        "GitHub Container Registry). If not specified, images will be stored at "
        "the registry root. For Docker Hub this would mean only official images "
        "can be pushed.",
    )

    @field_validator("uri")
    @classmethod
    def strip_trailing_slash(cls, uri: str) -> str:
        """Removes trailing slashes from the URI.

        Args:
            uri: The URI to be stripped.

        Returns:
            The URI without trailing slashes.
        """
        return uri.rstrip("/")

    @property
    def is_local(self) -> bool:
        """Checks if this stack component is running locally.

        Returns:
            True if this config is for a local component, False otherwise.
        """
        return bool(re.fullmatch(r"localhost:[0-9]{4,5}", self.uri))


class BaseContainerRegistry(AuthenticationMixin):
    """Base class for all ZenML container registries."""

    _docker_client: Optional["DockerClient"] = None

    @property
    def config(self) -> BaseContainerRegistryConfig:
        """Returns the `BaseContainerRegistryConfig` config.

        Returns:
            The configuration.
        """
        return cast(BaseContainerRegistryConfig, self._config)

    @property
    def requires_authentication(self) -> bool:
        """Returns whether the container registry requires authentication.

        Returns:
            `True` if the container registry requires authentication,
            `False` otherwise.
        """
        return bool(self.config.authentication_secret)

    @property
    def credentials(self) -> Optional[Tuple[str, str]]:
        """Username and password to authenticate with this container registry.

        Returns:
            Tuple with username and password if this container registry
            requires authentication, `None` otherwise.
        """
        secret = self.get_typed_authentication_secret(
            expected_schema_type=BasicAuthSecretSchema
        )
        if secret:
            return secret.username, secret.password

        connector = self.get_connector()
        if connector:
            from zenml.service_connectors.docker_service_connector import (
                DockerServiceConnector,
            )

            if isinstance(connector, DockerServiceConnector):
                return (
                    connector.config.username.get_secret_value(),
                    connector.config.password.get_secret_value(),
                )

        return None

    @property
    def docker_client(self) -> "DockerClient":
        """Returns a Docker client for this container registry.

        Returns:
            The Docker client.

        Raises:
            RuntimeError: If the connector does not return a Docker client.
        """
        from docker.client import DockerClient

        # Refresh the client also if the connector has expired
        if self._docker_client and not self.connector_has_expired():
            return self._docker_client

        connector = self.get_connector()
        if connector:
            client = connector.connect()
            if not isinstance(client, DockerClient):
                raise RuntimeError(
                    f"Expected a DockerClient while trying to use the "
                    f"linked connector, but got {type(client)}."
                )
            self._docker_client = client
        else:
            self._docker_client = (
                docker_utils._try_get_docker_client_from_env()
            )

            credentials = self.credentials
            if credentials:
                username, password = credentials
                self._docker_client.login(
                    username=username,
                    password=password,
                    registry=self.config.uri,
                    reauth=True,
                )

        return self._docker_client

    def is_valid_image_name_for_registry(self, image_name: str) -> bool:
        """Check if the image name is valid for the container registry.

        Args:
            image_name: The name of the image.

        Returns:
            `True` if the image name is valid for the container registry,
            `False` otherwise.
        """
        # Remove prefixes to make sure this logic also works for DockerHub
        image_name = image_name.removeprefix("index.docker.io/")
        image_name = image_name.removeprefix("docker.io/")

        registry_uri = self.config.uri.removeprefix("index.docker.io/")
        registry_uri = registry_uri.removeprefix("docker.io/")

        return image_name.startswith(registry_uri)

    def prepare_image_push(self, image_name: str) -> None:
        """Preparation before an image gets pushed.

        Subclasses can overwrite this to do any necessary checks or
        preparations before an image gets pushed.

        Args:
            image_name: Name of the docker image that will be pushed.
        """

    def push_image(self, image_name: str) -> str:
        """Pushes a docker image.

        Args:
            image_name: Name of the docker image that will be pushed.

        Returns:
            The Docker repository digest of the pushed image.

        Raises:
            ValueError: If the image name is not associated with this
                container registry.
        """
        if not self.is_valid_image_name_for_registry(image_name):
            raise ValueError(
                f"Docker image `{image_name}` does not belong to container "
                f"registry `{self.config.uri}`."
            )

        self.prepare_image_push(image_name)
        return docker_utils.push_image(
            image_name, docker_client=self.docker_client
        )

    def get_image_repo_digest(self, image_name: str) -> Optional[str]:
        """Get the repository digest of an image.

        Args:
            image_name: The name of the image.

        Returns:
            The repository digest of the image.
        """
        if not self.is_valid_image_name_for_registry(image_name):
            return None

        try:
            metadata = self.docker_client.images.get_registry_data(image_name)
        except Exception:
            return None

        return cast(str, metadata.id.split(":")[-1])


class BaseContainerRegistryFlavor(Flavor):
    """Base flavor for container registries."""

    @property
    def type(self) -> StackComponentType:
        """Returns the flavor type.

        Returns:
            The flavor type.
        """
        return StackComponentType.CONTAINER_REGISTRY

    @property
    def service_connector_requirements(
        self,
    ) -> Optional[ServiceConnectorRequirements]:
        """Service connector resource requirements for service connectors.

        Specifies resource requirements that are used to filter the available
        service connector types that are compatible with this flavor.

        Returns:
            Requirements for compatible service connectors, if a service
            connector is required for this flavor.
        """
        return ServiceConnectorRequirements(
            resource_type=DOCKER_REGISTRY_RESOURCE_TYPE,
            resource_id_attr="uri",
        )

    @property
    def config_class(self) -> Type[BaseContainerRegistryConfig]:
        """Config class for this flavor.

        Returns:
            The config class.
        """
        return BaseContainerRegistryConfig

    @property
    def implementation_class(self) -> Type[BaseContainerRegistry]:
        """Implementation class.

        Returns:
            The implementation class.
        """
        return BaseContainerRegistry



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Base class for ZenML step operators."""

import time
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Dict, List, Type, cast

from zenml.client import Client
from zenml.enums import ExecutionStatus, StackComponentType
from zenml.logger import get_logger
from zenml.models import StepRunResponse
from zenml.stack import Flavor, StackComponent
from zenml.stack.stack_component import StackComponentConfig
from zenml.step_operators.step_operator_entrypoint_configuration import (
    StepOperatorEntrypointConfiguration,
)

if TYPE_CHECKING:
    from zenml.config.step_run_info import StepRunInfo

logger = get_logger(__name__)


class BaseStepOperatorConfig(StackComponentConfig):
    """Base config for step operators."""


class BaseStepOperator(StackComponent, ABC):
    """Base class for all ZenML step operators."""

    @property
    def config(self) -> BaseStepOperatorConfig:
        """Returns the config of the step operator.

        Returns:
            The config of the step operator.
        """
        return cast(BaseStepOperatorConfig, self._config)

    @property
    def entrypoint_config_class(
        self,
    ) -> Type[StepOperatorEntrypointConfiguration]:
        """Returns the entrypoint configuration class for this step operator.

        Concrete step operator implementations may override this property
        to return a custom entrypoint configuration class if they need to
        customize the entrypoint configuration.

        Returns:
            The entrypoint configuration class for this step operator.
        """
        return StepOperatorEntrypointConfiguration

    def submit(
        self,
        info: "StepRunInfo",
        entrypoint_command: List[str],
        environment: Dict[str, str],
    ) -> None:
        """Submit a step run.

        This method should submit the step run and return without waiting for
        it to finish.

        Args:
            info: Information about the step run.
            entrypoint_command: Command that executes the step.
            environment: Environment variables to set in the step operator
                environment.

        Raises:
            NotImplementedError: If the step operator does not implement this
                method.
        """
        raise NotImplementedError(
            "Submitting step runs is not implemented for "
            f"the {self.__class__.__name__} step operator."
        )

    def get_status(self, step_run: "StepRunResponse") -> ExecutionStatus:
        """Get the status of a step run.

        Args:
            step_run: The step run to get the status of.

        Raises:
            NotImplementedError: If the step operator does not implement this
                method.
        """
        raise NotImplementedError(
            "Getting the status of step runs is not implemented for "
            f"the {self.__class__.__name__} step operator."
        )

    def wait(self, step_run: "StepRunResponse") -> ExecutionStatus:
        """Wait for a step run to finish.

        Args:
            step_run: The step run to wait for.

        Returns:
            The final status of the step run.
        """
        sleep_interval = 1
        max_sleep_interval = 16

        while True:
            try:
                status = self.get_status(step_run)
            except Exception as e:
                logger.error(
                    "Failed to get status of step run `%s`: %s",
                    step_run.id,
                    e,
                )
                # Fall back to the status of the ZenML server
                status = (
                    Client().get_run_step(step_run.id, hydrate=False).status
                )

            if status.is_finished or status == ExecutionStatus.RETRYING:
                return status

            logger.debug(
                "Waiting for step run with ID %s to finish (current "
                "status: %s)",
                step_run.id,
                status,
            )
            time.sleep(sleep_interval)
            if sleep_interval < max_sleep_interval:
                sleep_interval *= 2

    def cancel(self, step_run: "StepRunResponse") -> None:
        """Cancel a step run.

        Args:
            step_run: The step run to cancel.

        Raises:
            NotImplementedError: If the step operator does not implement this
                method.
        """
        raise NotImplementedError(
            "Canceling step runs is not implemented for "
            f"the {self.__class__.__name__} step operator."
        )


class BaseStepOperatorFlavor(Flavor):
    """Base class for all ZenML step operator flavors."""

    @property
    def type(self) -> StackComponentType:
        """Returns the flavor type.

        Returns:
            The type of the flavor.
        """
        return StackComponentType.STEP_OPERATOR

    @property
    def config_class(self) -> Type[BaseStepOperatorConfig]:
        """Returns the config class for this flavor.

        Returns:
            The config class for this flavor.
        """
        return BaseStepOperatorConfig

    @property
    @abstractmethod
    def implementation_class(self) -> Type[BaseStepOperator]:
        """Returns the implementation class for this flavor.

        Returns:
            The implementation class for this flavor.
        """



#  Copyright (c) ZenML GmbH 2023. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Pipeline configuration classes."""

import json
from typing import Any, Dict, List, Optional

from pydantic import Field

from zenml.config.frozen_base_model import FrozenBaseModel
from zenml.config.source import Source, SourceWithValidator
from zenml.config.step_configurations import StepSpec
from zenml.utils.json_utils import pydantic_encoder


class OutputSpec(FrozenBaseModel):
    """Pipeline output specification."""

    step_name: str
    output_name: str


class PipelineSpec(FrozenBaseModel):
    """Specification of a pipeline."""

    # Versions:
    # - 0.2: Legacy BasePipeline in release <=0.39.1, the upstream steps and
    #   inputs in the step specs refer to the step names, not the pipeline
    #   parameter names
    # - 0.3: Legacy BasePipeline in release >0.39.1, the upstream steps and
    #   inputs in the step specs refer to the pipeline parameter names
    # - 0.4: New Pipeline class, the upstream steps and
    #   inputs in the step specs refer to the pipeline parameter names
    # - 0.5: Adds input schema, outputs and output schema
    version: str = "0.5"
    source: Optional[SourceWithValidator] = None
    parameters: Dict[str, Any] = {}
    input_schema: Optional[Dict[str, Any]] = Field(
        default=None,
        description="JSON schema of the pipeline inputs. This is only set "
        "for pipeline specs with version >= 0.5. If the value is None, the "
        "schema generation failed, which is most likely because some of the "
        "pipeline inputs are not JSON serializable.",
    )
    steps: List[StepSpec]
    outputs: List[OutputSpec] = []
    output_schema: Optional[Dict[str, Any]] = Field(
        default=None,
        description="JSON schema of the pipeline outputs. This is only set "
        "for pipeline specs with version >= 0.5. If the value is None, the "
        "schema generation failed, which is most likely because some of the "
        "pipeline outputs are not JSON serializable.",
    )

    def __eq__(self, other: Any) -> bool:
        """Returns whether the other object is referring to the same pipeline.

        Args:
            other: The other object to compare to.

        Returns:
            True if the other object is referring to the same pipeline.
        """
        if isinstance(other, PipelineSpec):
            return self.steps == other.steps
        return NotImplemented

    @property
    def json_with_string_sources(self) -> str:
        """JSON representation with sources replaced by their import path.

        Returns:
            The JSON representation.
        """
        from packaging import version

        dict_ = self.model_dump()

        if self.source:
            dict_["source"] = self.source.import_path

        for step_dict in dict_["steps"]:
            step_dict["source"] = Source.model_validate(
                step_dict["source"]
            ).import_path

        if version.parse(self.version) < version.parse("0.4"):
            # Keep backwards compatibility with old pipeline versions
            dict_.pop("source")
            dict_.pop("parameters")

        return json.dumps(dict_, sort_keys=False, default=pydantic_encoder)



#  Copyright (c) ZenML GmbH 2023. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Implementation of the GCP Secrets Store."""

import json
import os
from typing import (
    Any,
    ClassVar,
    Dict,
    Optional,
    Type,
    cast,
)
from uuid import UUID

from google.api_core import exceptions as google_exceptions
from google.cloud.secretmanager import SecretManagerServiceClient
from pydantic import ConfigDict, model_validator

from zenml.enums import (
    SecretsStoreType,
)
from zenml.integrations.gcp import (
    GCP_CONNECTOR_TYPE,
    GCP_RESOURCE_TYPE,
)
from zenml.integrations.gcp.service_connectors.gcp_service_connector import (
    GCPAuthenticationMethods,
)
from zenml.logger import get_logger
from zenml.utils.pydantic_utils import before_validator_handler
from zenml.zen_stores.secrets_stores.service_connector_secrets_store import (
    ServiceConnectorSecretsStore,
    ServiceConnectorSecretsStoreConfiguration,
)

logger = get_logger(__name__)


GCP_ZENML_SECRET_NAME_PREFIX = "zenml"
ZENML_GROUP_KEY = "zenml-group-key"
ZENML_GCP_SECRET_SCOPE_PATH_SEPARATOR = "-"


class GCPSecretsStoreConfiguration(ServiceConnectorSecretsStoreConfiguration):
    """GCP secrets store configuration.

    Attributes:
        type: The type of the store.
    """

    type: SecretsStoreType = SecretsStoreType.GCP

    @property
    def project_id(self) -> str:
        """Get the GCP project ID.

        Returns:
            The GCP project ID.

        Raises:
            ValueError: If the project ID is not set.
        """
        project_id = self.auth_config.get("project_id")
        if project_id:
            return str(project_id)

        raise ValueError("GCP `project_id` must be specified in auth_config.")

    @model_validator(mode="before")
    @classmethod
    @before_validator_handler
    def populate_config(cls, data: Dict[str, Any]) -> Dict[str, Any]:
        """Populate the connector configuration from legacy attributes.

        Args:
            data: Dict representing user-specified runtime settings.

        Returns:
            Validated settings.
        """
        # Search for legacy attributes and populate the connector configuration
        # from them, if they exist.
        if data.get("project_id"):
            if not os.environ.get("GOOGLE_APPLICATION_CREDENTIALS"):
                logger.warning(
                    "The `project_id` GCP secrets store attribute is "
                    "deprecated and will be removed in a future version of ZenML. "
                    "Please use the `auth_method` and `auth_config` attributes "
                    "instead. Using an implicit GCP authentication to access "
                    "the GCP Secrets Manager API."
                )
                data["auth_method"] = GCPAuthenticationMethods.IMPLICIT
                data["auth_config"] = dict(
                    project_id=data.get("project_id"),
                )
            else:
                logger.warning(
                    "The `project_id` GCP secrets store attribute and the "
                    "`GOOGLE_APPLICATION_CREDENTIALS` environment variable are "
                    "deprecated and will be removed in a future version of ZenML. "
                    "Please use the `auth_method` and `auth_config` attributes "
                    "instead."
                )
                data["auth_method"] = GCPAuthenticationMethods.SERVICE_ACCOUNT
                data["auth_config"] = dict(
                    project_id=data.get("project_id"),
                )
                # Load the service account credentials from the file
                with open(os.environ["GOOGLE_APPLICATION_CREDENTIALS"]) as f:
                    data["auth_config"]["service_account_json"] = f.read()

        return data

    model_config = ConfigDict(extra="allow")


class GCPSecretsStore(ServiceConnectorSecretsStore):
    """Secrets store implementation that uses the GCP Secrets Manager API."""

    config: GCPSecretsStoreConfiguration
    TYPE: ClassVar[SecretsStoreType] = SecretsStoreType.GCP
    CONFIG_TYPE: ClassVar[Type[ServiceConnectorSecretsStoreConfiguration]] = (
        GCPSecretsStoreConfiguration
    )
    SERVICE_CONNECTOR_TYPE: ClassVar[str] = GCP_CONNECTOR_TYPE
    SERVICE_CONNECTOR_RESOURCE_TYPE: ClassVar[str] = GCP_RESOURCE_TYPE

    _client: Optional[SecretManagerServiceClient] = None

    @property
    def client(self) -> SecretManagerServiceClient:
        """Initialize and return the GCP Secrets Manager client.

        Returns:
            The GCP Secrets Manager client instance.
        """
        return cast(SecretManagerServiceClient, super().client)

    # ====================================
    # Secrets Store interface implementation
    # ====================================

    # --------------------------------
    # Initialization and configuration
    # --------------------------------

    def _initialize_client_from_connector(self, client: Any) -> Any:
        """Initialize the GCP Secrets Manager client from the service connector client.

        Args:
            client: The authenticated client object returned by the service
                connector.

        Returns:
            The GCP Secrets Manager client.
        """
        return SecretManagerServiceClient(credentials=client)

    # ------
    # Secrets
    # ------

    @property
    def parent_name(self) -> str:
        """Construct the GCP parent path to the secret manager.

        Returns:
            The parent path to the secret manager
        """
        return f"projects/{self.config.project_id}"

    def _get_gcp_secret_name(
        self,
        secret_id: UUID,
    ) -> str:
        """Get the GCP secret name for the given secret.

        The convention used for GCP secret names is to use the ZenML
        secret UUID prefixed with `zenml` as the AWS secret name,
        i.e. `zenml/<secret_uuid>`.

        Args:
            secret_id: The ZenML secret ID.

        Returns:
            The GCP secret name.
        """
        return f"{GCP_ZENML_SECRET_NAME_PREFIX}-{str(secret_id)}"

    def store_secret_values(
        self,
        secret_id: UUID,
        secret_values: Dict[str, str],
    ) -> None:
        """Store secret values for a new secret.

        Args:
            secret_id: ID of the secret.
            secret_values: Values for the secret.

        Raises:
            RuntimeError: if the GCP Secrets Manager API returns an unexpected
                error.
        """
        secret_value = json.dumps(secret_values)

        labels = self._get_secret_metadata(secret_id=secret_id)

        try:
            gcp_secret = self.client.create_secret(
                request={
                    "parent": self.parent_name,
                    "secret_id": self._get_gcp_secret_name(secret_id),
                    "secret": {
                        "replication": {"automatic": {}},
                        "labels": labels,
                    },
                }
            )

            logger.debug(f"Created empty GCP parent secret: {gcp_secret.name}")

            self.client.add_secret_version(
                request={
                    "parent": gcp_secret.name,
                    "payload": {"data": secret_value.encode()},
                }
            )

            logger.debug(f"Added value to GCP secret {gcp_secret.name}")
        except Exception as e:
            raise RuntimeError(f"Failed to create secret.: {str(e)}") from e

        logger.debug(f"Created GCP secret {gcp_secret.name}")

    def get_secret_values(self, secret_id: UUID) -> Dict[str, str]:
        """Get the secret values for an existing secret.

        Args:
            secret_id: ID of the secret.

        Returns:
            The secret values.

        Raises:
            KeyError: if no secret values for the given ID are stored in the
                secrets store.
            RuntimeError: if the GCP Secrets Manager API returns an unexpected
                error.
        """
        gcp_secret_name = self.client.secret_path(
            self.config.project_id,
            self._get_gcp_secret_name(secret_id=secret_id),
        )

        try:
            secret = self.client.get_secret(name=gcp_secret_name)
            secret_version_values = self.client.access_secret_version(
                name=f"{gcp_secret_name}/versions/latest"
            )
        except google_exceptions.NotFound as e:
            raise KeyError(
                f"Can't find the secret values for secret ID '{secret_id}' "
                f"in the secrets store back-end: {str(e)}"
            ) from e
        except Exception as e:
            raise RuntimeError(
                f"Error fetching secret with ID {secret_id} {e}"
            )

        # The GCP secret labels do not really behave like a dictionary: when
        # a key is not found, it does not raise a KeyError, but instead
        # returns an empty string. That's why we make this conversion.
        metadata = dict(secret.labels)

        # The _verify_secret_metadata method raises a KeyError if the
        # secret is not valid or does not belong to this server. Here we
        # simply pass the exception up the stack, as if the secret was not found
        # in the first place.
        self._verify_secret_metadata(
            secret_id=secret_id,
            metadata=metadata,
        )

        secret_values = json.loads(
            secret_version_values.payload.data.decode("UTF-8")
        )

        if not isinstance(secret_values, dict):
            raise RuntimeError(
                f"Google secret values for secret ID {gcp_secret_name} could "
                "not be decoded: expected a dictionary."
            )

        logger.debug(f"Fetched GCP secret: {gcp_secret_name}")

        return secret_values

    def update_secret_values(
        self,
        secret_id: UUID,
        secret_values: Dict[str, str],
    ) -> None:
        """Updates secret values for an existing secret.

        Args:
            secret_id: The ID of the secret to be updated.
            secret_values: The new secret values.

        Raises:
            RuntimeError: if the GCP Secrets Manager API returns an unexpected
                error.
        """
        gcp_secret_name = self.client.secret_path(
            self.config.project_id,
            self._get_gcp_secret_name(secret_id=secret_id),
        )

        # Convert the ZenML secret metadata to GCP labels
        metadata = self._get_secret_metadata(secret_id)

        try:
            # Update the secret metadata
            update_secret = {
                "name": gcp_secret_name,
                "labels": metadata,
            }
            update_mask = {"paths": ["labels"]}
            gcp_updated_secret = self.client.update_secret(
                request={
                    "secret": update_secret,
                    "update_mask": update_mask,
                }
            )
            # Add a new secret version
            secret_value = json.dumps(secret_values)
            self.client.add_secret_version(
                request={
                    "parent": gcp_updated_secret.name,
                    "payload": {"data": secret_value.encode()},
                }
            )
        except Exception as e:
            raise RuntimeError(f"Error updating secret: {e}") from e

        logger.debug(f"Updated GCP secret: {gcp_secret_name}")

    def delete_secret_values(self, secret_id: UUID) -> None:
        """Deletes secret values for an existing secret.

        Args:
            secret_id: The ID of the secret.

        Raises:
            KeyError: if no secret values for the given ID are stored in the
                secrets store.
            RuntimeError: if the GCP Secrets Manager API returns an unexpected
                error.
        """
        gcp_secret_name = self.client.secret_path(
            self.config.project_id,
            self._get_gcp_secret_name(secret_id=secret_id),
        )

        try:
            self.client.delete_secret(request={"name": gcp_secret_name})
        except google_exceptions.NotFound:
            raise KeyError(f"Secret with ID {secret_id} not found")
        except Exception as e:
            raise RuntimeError(f"Failed to delete secret: {str(e)}") from e

        logger.debug(f"Deleted GCP secret: {gcp_secret_name}")



#  Copyright (c) ZenML GmbH 2023. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Utility functions for handling environment variables."""

import contextlib
import os
import re
from abc import ABC, abstractmethod
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterator,
    List,
    Match,
    Optional,
    TypeVar,
    Union,
    cast,
)

from pydantic import BaseModel
from typing_extensions import Self

from zenml.client import Client
from zenml.logger import get_logger
from zenml.utils import string_utils

if TYPE_CHECKING:
    from zenml.config.pipeline_configurations import PipelineConfiguration
    from zenml.config.step_configurations import StepConfiguration
    from zenml.models import StackResponse
    from zenml.stack import Stack

logger = get_logger(__name__)

V = TypeVar("V", bound=Any)
ENV_VARIABLE_PLACEHOLDER_PATTERN = re.compile(pattern=r"\$\{([a-zA-Z0-9_]+)\}")

ENV_VAR_CHUNK_SUFFIX = "_CHUNK_"


def split_environment_variables(
    size_limit: int,
    env: Optional[Dict[str, str]] = None,
) -> None:
    """Split long environment variables into chunks.

    Splits the input environment variables with values that exceed the supplied
    maximum length into individual components. The input environment variables
    are modified in-place.

    Args:
        size_limit: Maximum length of an environment variable value.
        env: Input environment variables dictionary. If not supplied, the
            OS environment variables are used.

    Raises:
        RuntimeError: If an environment variable value is too large and requires
            more than 10 chunks.
    """
    if env is None:
        env = cast(Dict[str, str], os.environ)

    for key, value in env.copy().items():
        if len(value) <= size_limit:
            continue

        # We keep the number of chunks to a maximum of 10 to avoid generating
        # too many environment variables chunks and also to make the
        # reconstruction easier to implement
        if len(value) > size_limit * 10:
            raise RuntimeError(
                f"Environment variable {key} exceeds the maximum length of "
                f"{size_limit * 10} characters."
            )

        env.pop(key)

        # Split the environment variable into chunks
        chunks = [
            value[i : i + size_limit] for i in range(0, len(value), size_limit)
        ]
        for i, chunk in enumerate(chunks):
            env[f"{key}{ENV_VAR_CHUNK_SUFFIX}{i}"] = chunk


def reconstruct_environment_variables(
    env: Optional[Dict[str, str]] = None,
) -> None:
    """Reconstruct environment variables that were split into chunks.

    Reconstructs the environment variables with values that were split into
    individual chunks because they were too large. The input environment
    variables are modified in-place.

    Args:
        env: Input environment variables dictionary. If not supplied, the OS
            environment variables are used.
    """
    if env is None:
        env = cast(Dict[str, str], os.environ)

    chunks: Dict[str, List[str]] = {}
    for key in env.keys():
        if not key[:-1].endswith(ENV_VAR_CHUNK_SUFFIX):
            continue

        # Collect all chunks of the same environment variable
        original_key = key[: -(len(ENV_VAR_CHUNK_SUFFIX) + 1)]
        chunks.setdefault(original_key, [])
        chunks[original_key].append(key)

    # Reconstruct the environment variables from their chunks
    for key, chunk_keys in chunks.items():
        chunk_keys.sort()
        value = "".join([env[key] for key in chunk_keys])
        env[key] = value

        # Remove the chunk environment variables
        for key in chunk_keys:
            env.pop(key)


def substitute_env_variable_placeholders(
    value: V, raise_when_missing: bool = True
) -> V:
    """Substitute environment variable placeholders in an object.

    Args:
        value: The object in which to substitute the placeholders.
        raise_when_missing: If True, an exception will be raised when an
            environment variable is missing. Otherwise, a warning will be logged
            instead.

    Returns:
        The object with placeholders substituted.
    """

    def _replace_with_env_variable_value(match: Match[str]) -> str:
        key = match.group(1)
        if key in os.environ:
            return os.environ[key]
        else:
            if raise_when_missing:
                raise KeyError(
                    "Unable to substitute environment variable placeholder "
                    f"'{key}' because the environment variable is not set."
                )
            else:
                logger.warning(
                    "Unable to substitute environment variable placeholder %s "
                    "because the environment variable is not set, using an "
                    "empty string instead.",
                    key,
                )
                return ""

    def _substitution_func(v: str) -> str:
        return ENV_VARIABLE_PLACEHOLDER_PATTERN.sub(
            _replace_with_env_variable_value, v
        )

    return string_utils.substitute_string(
        value=value, substitution_func=_substitution_func
    )


@contextlib.contextmanager
def temporary_environment(environment: Dict[str, str]) -> Iterator[None]:
    """Temporarily set environment variables.

    Args:
        environment: The environment variables to set.

    Yields:
        Nothing.
    """
    try:
        previous_env = {}
        for key, value in environment.items():
            previous_env[key] = os.environ.get(key, None)
            os.environ[key] = value
        yield
    finally:
        for key, previous_value in previous_env.items():
            updated_value = environment[key]
            current_value = os.environ.get(key)

            if current_value != updated_value:
                # The environment variable got updated while this context
                # manager was active -> We don't reset it back to the old
                # value
                continue
            else:
                if previous_value is None:
                    os.environ.pop(key, None)
                else:
                    os.environ[key] = previous_value


def get_runtime_environment(
    config: Union["PipelineConfiguration", "StepConfiguration"],
    stack: Union["Stack", "StackResponse"],
) -> Dict[str, str]:
    """Get the runtime environment variables.

    Args:
        config: The configuration.
        stack: The stack on which the pipeline/step will run.

    Returns:
        A dictionary of environment variables.
    """
    environment = {}
    for component in stack.components.values():
        if isinstance(component, list):
            for item in component:
                environment.update(item.environment)
        else:
            environment.update(component.environment)

    environment.update(stack.environment)
    environment.update(config.environment)

    for key, value in environment.items():
        environment[key] = str(value)

    return environment


def get_runtime_secret_environment(
    config: Union["PipelineConfiguration", "StepConfiguration"],
    stack: Union["Stack", "StackResponse"],
) -> Dict[str, str]:
    """Get the runtime secret environment variables.

    Args:
        config: The configuration.
        stack: The stack on which the pipeline/step will run.

    Returns:
        A dictionary of environment variables.
    """
    # We reverse the secrets to make sure the last defined secrets
    # override previous ones.
    # In the case of a step, the secrets will be the pipeline secrets followed
    # by the step secrets.
    secrets = list(reversed(config.secrets))
    secrets.extend(stack.secrets)

    for component in stack.components.values():
        if isinstance(component, list):
            for item in component:
                secrets.extend(item.secrets)
        else:
            secrets.extend(component.secrets)

    # Removes duplicates while preserving order, only the first occurrence of
    # each secret will be kept. We then reverse the list to make sure the
    # overrides are applied in the correct order.
    secrets = list(reversed(dict.fromkeys(secrets)))

    environment = {}
    for secret_name_or_id in secrets:
        try:
            secret = Client().get_secret(secret_name_or_id)
        except Exception as e:
            logger.warning(
                "Failed to get secret `%s` with error: %s. Skipping setting "
                "environment variables for this secret.",
                secret_name_or_id,
                e,
            )
            continue

        if not secret.secret_values:
            logger.warning(
                "Did not find any secret values for secret `%s`. This might be "
                "because you do not have permissions to read the secret "
                "values. Skipping setting environment variables for this "
                "secret.",
                secret_name_or_id,
            )
            continue

        for key, value in secret.secret_values.items():
            environment[key] = str(value)

    return environment


@contextlib.contextmanager
def temporary_runtime_environment(
    config: Union["PipelineConfiguration", "StepConfiguration"],
    stack: Union["Stack", "StackResponse"],
) -> Iterator[None]:
    """Temporarily set runtime environment variables.

    Args:
        config: The configuration.
        stack: The stack on which the pipeline/step will run.

    Yields:
        Nothing.
    """
    env = get_runtime_environment(config, stack)
    secret_env = get_runtime_secret_environment(config, stack)
    env.update(secret_env)

    with temporary_environment(env):
        yield


class ConfigBase(BaseModel, ABC):
    """Base class for configuration objects. Useful for config inheritance."""

    @staticmethod
    @abstractmethod
    def prefixes() -> list[str]:
        """Prefixes to use for env var name resolution.

        Returns:
            A list of prefixes to use for env var name resolution.
        """
        raise NotImplementedError

    @classmethod
    def get_env_vars(cls) -> dict[str, str]:
        """Method resolving the class variables from the environment.

        Notes:
        - Beware of variable name conflicts when composing configs via
        inheritance. For instance, if you have DB and Cache env vars with a
        different prefix (e.g. {ZENML_REDIS_}CONN_URL, {ZENML_DB_}CONN_URL) this
        may lead to a conflict in naming when combining the 2 individual config
        objects (We get 2 variables named conn_url).
        - To avoid generating ambiguous configs the function detects duplicates and fails eagerly.
        You can avoid such behaviors by using more generic prefixes and more precise
        variable names (e.g. prefix = ZENML_, variable = db_conn_url) or by controlling
        inheritance (try not to combine too many config classes, especially with common variable names).
        - Why not use pydantic_settings.BaseSettings? Because - while neat - it doesn't provide the
        ability to compose config objects with different prefixes via inheritance.

        Returns:
            A dictionary of config variables.

        Raises:
            RuntimeError: If config has conflicting variable names.
        """
        sorted_prefixes = sorted(
            set(cls.prefixes()), key=lambda p: (-len(p), p)
        )
        logger.info("Config resolving prefixes: %s", sorted_prefixes)

        env_vars: dict[str, str] = os.environ.copy()
        parameters: dict[str, str] = {}
        sources: dict[str, str] = {}  # variable_name -> env key

        for prefix in sorted_prefixes:
            found_keys: set[str] = set()

            for key, value in env_vars.items():
                if not key.startswith(prefix):
                    continue

                found_keys.add(key)
                variable_name = key.removeprefix(prefix).lower()

                if variable_name in parameters:
                    raise RuntimeError(
                        f"Environment variable extraction failed for {cls.__name__}. "
                        f"Variable '{variable_name}' is not uniquely defined; it maps to both "
                        f"{sources[variable_name]!r} and {key!r}."
                    )

                parameters[variable_name] = value
                sources[variable_name] = key

            env_vars = {
                k: v for k, v in env_vars.items() if k not in found_keys
            }

        return parameters

    @classmethod
    def load_from_env(cls) -> Self:
        """Resolve the environment and instantiate a new instance of the class.

        Returns:
            An instance of the class.
        """
        env_vars = cls.get_env_vars()
        logger.info(
            "Config identified the following variables: %s",
            list(env_vars.keys()),
        )
        return cls(**env_vars)



#  Copyright (c) ZenML GmbH 2024. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Implementation of the AzureML Orchestrator."""

import json
import os
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    Optional,
    Tuple,
    Type,
    Union,
    cast,
)
from uuid import UUID

from azure.ai.ml import Input, MLClient, Output, command
from azure.ai.ml.constants import TimeZone
from azure.ai.ml.dsl import pipeline
from azure.ai.ml.entities import (
    CommandComponent,
    CronTrigger,
    Environment,
    JobResourceConfiguration,
    JobSchedule,
    RecurrenceTrigger,
)
from azure.core.exceptions import (
    HttpResponseError,
    ResourceExistsError,
)
from azure.identity import DefaultAzureCredential

from zenml.config.base_settings import BaseSettings
from zenml.config.step_configurations import Step
from zenml.constants import (
    METADATA_ORCHESTRATOR_RUN_ID,
    METADATA_ORCHESTRATOR_URL,
    ORCHESTRATOR_DOCKER_IMAGE_KEY,
)
from zenml.enums import ExecutionStatus, StackComponentType
from zenml.integrations.azure.azureml_utils import (
    convert_job_status,
    create_or_get_compute,
)
from zenml.integrations.azure.flavors.azureml import AzureMLComputeTypes
from zenml.integrations.azure.flavors.azureml_orchestrator_flavor import (
    AzureMLOrchestratorConfig,
    AzureMLOrchestratorSettings,
)
from zenml.integrations.azure.orchestrators.azureml_orchestrator_entrypoint_config import (
    AzureMLEntrypointConfiguration,
)
from zenml.logger import get_logger
from zenml.metadata.metadata_types import MetadataType, Uri
from zenml.orchestrators import ContainerizedOrchestrator, SubmissionResult
from zenml.orchestrators.publish_utils import publish_step_run_metadata
from zenml.orchestrators.utils import get_orchestrator_run_name
from zenml.pipelines.dynamic.entrypoint_configuration import (
    DynamicPipelineEntrypointConfiguration,
)
from zenml.stack import StackValidator
from zenml.step_operators.step_operator_entrypoint_configuration import (
    StepOperatorEntrypointConfiguration,
)
from zenml.utils.string_utils import b64_encode

if TYPE_CHECKING:
    from zenml.config.step_run_info import StepRunInfo
    from zenml.models import (
        PipelineRunResponse,
        PipelineSnapshotResponse,
        StepRunResponse,
    )
    from zenml.stack import Stack

logger = get_logger(__name__)

ENV_ZENML_AZUREML_RUN_ID = "AZUREML_ROOT_RUN_ID"
STEP_JOB_NAME_METADATA_KEY = "job_name"


class AzureMLOrchestrator(ContainerizedOrchestrator):
    """Orchestrator responsible for running pipelines on AzureML."""

    _azureml_client: Optional[MLClient] = None

    @property
    def config(self) -> AzureMLOrchestratorConfig:
        """Returns the `AzureMLOrchestratorConfig` config.

        Returns:
            The configuration.
        """
        return cast(AzureMLOrchestratorConfig, self._config)

    @property
    def settings_class(self) -> Optional[Type["BaseSettings"]]:
        """Settings class for the AzureML orchestrator.

        Returns:
            The settings class.
        """
        return AzureMLOrchestratorSettings

    @property
    def validator(self) -> Optional[StackValidator]:
        """Validates the stack.

        In the remote case, checks that the stack contains a container registry,
        image builder and only remote components.

        Returns:
            A `StackValidator` instance.
        """

        def _validate_remote_components(
            stack: "Stack",
        ) -> Tuple[bool, str]:
            for component in stack.components.values():
                if not component.config.is_local:
                    continue

                return False, (
                    f"The AzureML orchestrator runs pipelines remotely, "
                    f"but the '{component.name}' {component.type.value} is "
                    "a local stack component and will not be available in "
                    "the AzureML step.\nPlease ensure that you always "
                    "use non-local stack components with the AzureML "
                    "orchestrator."
                )

            return True, ""

        return StackValidator(
            required_components={
                StackComponentType.CONTAINER_REGISTRY,
                StackComponentType.IMAGE_BUILDER,
            },
            custom_validation_function=_validate_remote_components,
        )

    def get_orchestrator_run_id(self) -> str:
        """Returns the run id of the active orchestrator run.

        Important: This needs to be a unique ID and return the same value for
        all steps of a pipeline run.

        Returns:
            The orchestrator run id.

        Raises:
            RuntimeError: If the run id cannot be read from the environment.
        """
        for env_var in [
            ENV_ZENML_AZUREML_RUN_ID,
            # Best-effort fallbacks for AzureML job runtimes
            "AZUREML_RUN_ID",
            "AZUREML_JOB_ID",
        ]:
            if env_var in os.environ:
                return os.environ[env_var]

        raise RuntimeError(
            "Unable to read orchestrator run id from environment."
        )

    @staticmethod
    def _create_command_component(
        step: Step,
        step_name: str,
        env_name: str,
        image: str,
        command: List[str],
        arguments: List[str],
        shm_size: Optional[str] = None,
    ) -> CommandComponent:
        """Creates a CommandComponent to run on AzureML Pipelines.

        Args:
            step: The step definition in ZenML.
            step_name: The name of the step.
            env_name: The name of the environment.
            image: The image to use in the environment
            command: The command to execute the entrypoint with.
            arguments: The arguments to pass into the command.
            shm_size: Optional shared memory size for the container (e.g., '2g', '200g').

        Returns:
            the generated AzureML CommandComponent.
        """
        env = Environment(name=env_name, image=image)

        outputs = {"completed": Output(type="uri_file")}

        inputs = {}
        if step.spec.upstream_steps:
            inputs = {
                f"{upstream_step}": Input(type="uri_file")
                for upstream_step in step.spec.upstream_steps
            }

        resources = None
        if shm_size:
            resources = JobResourceConfiguration(shm_size=shm_size)

        return CommandComponent(
            name=step_name,
            display_name=step_name,
            description=f"AzureML CommandComponent for {step_name}.",
            inputs=inputs,
            outputs=outputs,
            environment=env,
            command=" ".join(command + arguments),
            resources=resources,
        )

    def get_azureml_client(self) -> MLClient:
        """Returns the AzureML client.

        Returns:
            The AzureML client.
        """
        if self.connector_has_expired():
            self._azureml_client = None

        if self._azureml_client is None:
            if connector := self.get_connector():
                credentials = connector.connect()
            else:
                credentials = DefaultAzureCredential()

            self._azureml_client = MLClient(
                credential=credentials,
                subscription_id=self.config.subscription_id,
                resource_group_name=self.config.resource_group,
                workspace_name=self.config.workspace,
            )

        return self._azureml_client

    def submit_pipeline(
        self,
        snapshot: "PipelineSnapshotResponse",
        stack: "Stack",
        base_environment: Dict[str, str],
        step_environments: Dict[str, Dict[str, str]],
        placeholder_run: Optional["PipelineRunResponse"] = None,
    ) -> Optional[SubmissionResult]:
        """Submits a pipeline to the orchestrator.

        This method should only submit the pipeline and not wait for it to
        complete. If the orchestrator is configured to wait for the pipeline run
        to complete, a function that waits for the pipeline run to complete can
        be passed as part of the submission result.

        Args:
            snapshot: The pipeline snapshot to submit.
            stack: The stack the pipeline will run on.
            base_environment: Base environment shared by all steps. This should
                be set if your orchestrator for example runs one container that
                is responsible for starting all the steps.
            step_environments: Environment variables to set when executing
                specific steps.
            placeholder_run: An optional placeholder run for the snapshot.

        Raises:
            RuntimeError: If the creation of the schedule fails.

        Returns:
            Optional submission result.
        """
        # Settings
        settings = cast(
            AzureMLOrchestratorSettings,
            self.get_settings(snapshot),
        )

        # Client creation
        ml_client = self.get_azureml_client()

        # Create components
        components = {}
        for step_name, step in snapshot.step_configurations.items():
            step_environment = step_environments[step_name]
            # Get the image for each step
            image = self.get_image(snapshot=snapshot, step_name=step_name)

            # Get the command and arguments
            command = AzureMLEntrypointConfiguration.get_entrypoint_command()
            arguments = (
                AzureMLEntrypointConfiguration.get_entrypoint_arguments(
                    step_name=step_name,
                    snapshot_id=snapshot.id,
                    zenml_env_variables=b64_encode(
                        json.dumps(step_environment)
                    ),
                )
            )

            # Generate an AzureML CommandComponent
            components[step_name] = self._create_command_component(
                step=step,
                step_name=step_name,
                env_name=snapshot.pipeline_configuration.name,
                image=image,
                command=command,
                arguments=arguments,
                shm_size=settings.shm_size,
            )

        # Pipeline definition
        pipeline_args = dict()
        run_name = get_orchestrator_run_name(
            pipeline_name=snapshot.pipeline_configuration.name
        )
        pipeline_args["name"] = run_name

        if compute_target := create_or_get_compute(
            ml_client, settings, default_compute_name=f"zenml_{self.id}"
        ):
            pipeline_args["compute"] = compute_target

        @pipeline(force_rerun=True, **pipeline_args)  # type: ignore[call-overload, misc]
        def azureml_pipeline() -> None:
            """Create an AzureML pipeline."""
            # Here we have to track the inputs and outputs so that we can bind
            # the components to each other to execute them in a specific order.
            component_outputs: Dict[str, Any] = {}
            for component_name, component in components.items():
                # Inputs
                component_inputs = {}
                if component.inputs:
                    component_inputs.update(
                        {i: component_outputs[i] for i in component.inputs}
                    )

                # Job
                component_job = component(**component_inputs)

                # Apply per-step compute settings if available
                step_config = snapshot.step_configurations[component_name]
                step_settings = self.get_settings(step_config)

                if step_settings and isinstance(
                    step_settings, AzureMLOrchestratorSettings
                ):
                    # Check if step settings differ from pipeline settings to avoid unnecessary Azure calls
                    if (
                        step_settings.mode != settings.mode
                        or step_settings.compute_name != settings.compute_name
                        or step_settings.size != settings.size
                    ):
                        try:
                            step_compute = create_or_get_compute(
                                ml_client,
                                step_settings,
                                default_compute_name="zenml_{}_{}".format(
                                    self.id, component_name
                                ),
                            )
                            if step_compute:
                                component_job.compute = step_compute
                                logger.info(
                                    "Step '%s' will run on compute: %s",
                                    component_name,
                                    step_compute,
                                )
                        except Exception as e:
                            logger.debug(
                                "Could not apply per-step compute for '%s': %s",
                                component_name,
                                e,
                            )
                    else:
                        logger.debug(
                            "Step '%s' compute settings match pipeline level, using default compute",
                            component_name,
                        )

                # Outputs
                if component_job.outputs:
                    component_outputs[component_name] = (
                        component_job.outputs.completed
                    )

        # Create and execute the pipeline job
        pipeline_job = azureml_pipeline()

        if settings.mode == AzureMLComputeTypes.SERVERLESS:
            pipeline_job.settings.default_compute = "serverless"

        # Scheduling
        if schedule := snapshot.schedule:
            try:
                schedule_trigger: Optional[
                    Union[CronTrigger, RecurrenceTrigger]
                ] = None

                start_time = None
                if schedule.start_time is not None:
                    start_time = schedule.start_time.isoformat()

                end_time = None
                if schedule.end_time is not None:
                    end_time = schedule.end_time.isoformat()

                if schedule.cron_expression:
                    # If we are working with a cron expression
                    schedule_trigger = CronTrigger(
                        expression=schedule.cron_expression,
                        start_time=start_time,
                        end_time=end_time,
                        time_zone=TimeZone.UTC,
                    )

                elif schedule.interval_second:
                    # If we are working with intervals
                    interval = schedule.interval_second.total_seconds()

                    if interval % 60 != 0:
                        logger.warning(
                            "The ZenML AzureML orchestrator only works with "
                            "time intervals defined over minutes. Will "
                            f"use a schedule over {int(interval // 60)}."
                        )

                    if interval < 60:
                        raise RuntimeError(
                            "Can not create a schedule with an interval less "
                            "than 60 secs."
                        )

                    frequency = "minute"
                    interval = int(interval // 60)

                    schedule_trigger = RecurrenceTrigger(
                        frequency=frequency,
                        interval=interval,
                        start_time=start_time,
                        end_time=end_time,
                        time_zone=TimeZone.UTC,
                    )

                if schedule_trigger:
                    # Create and execute the job schedule
                    job_schedule = JobSchedule(
                        name=run_name,
                        trigger=schedule_trigger,
                        create_job=pipeline_job,
                    )
                    ml_client.schedules.begin_create_or_update(
                        job_schedule
                    ).result()
                    logger.info(
                        f"Scheduled pipeline '{run_name}' with recurrence "
                        "or cron expression."
                    )
                else:
                    raise RuntimeError(
                        "No valid scheduling configuration found for "
                        f"pipeline '{run_name}'."
                    )

            except (HttpResponseError, ResourceExistsError) as e:
                raise RuntimeError(
                    "Failed to create schedule for the pipeline "
                    f"'{run_name}': {str(e)}"
                )
            return None
        else:
            job = ml_client.jobs.create_or_update(pipeline_job)
            logger.info(f"Pipeline {run_name} has been started.")

            assert job.services is not None
            assert job.name is not None

            logger.info(
                f"Pipeline {run_name} is running. "
                "You can view the pipeline in the AzureML portal at "
                f"{job.services['Studio'].endpoint}"
            )

            _wait_for_completion = None
            if settings.synchronous:

                def _wait_for_completion() -> None:
                    logger.info("Waiting for pipeline to finish...")
                    ml_client.jobs.stream(job.name)

            return SubmissionResult(
                metadata=self.compute_metadata(job),
                wait_for_completion=_wait_for_completion,
            )

    def submit_dynamic_pipeline(
        self,
        snapshot: "PipelineSnapshotResponse",
        stack: "Stack",
        environment: Dict[str, str],
        placeholder_run: Optional["PipelineRunResponse"] = None,
    ) -> Optional[SubmissionResult]:
        """Submits a dynamic pipeline to the orchestrator.

        Dynamic pipelines are executed as an AzureML command job that runs the
        dynamic pipeline orchestration container.

        Args:
            snapshot: The pipeline snapshot to submit.
            stack: The stack the pipeline will run on.
            environment: Environment variables to set in the orchestration
                environment.
            placeholder_run: An optional placeholder run.

        Raises:
            RuntimeError: If the snapshot contains a schedule.

        Returns:
            Optional submission result.
        """
        logger.info("Submitting dynamic pipeline to AzureML...")

        if snapshot.schedule:
            raise RuntimeError(
                "Scheduling dynamic pipelines is not supported for the "
                "AzureML orchestrator yet."
            )

        settings = cast(
            AzureMLOrchestratorSettings,
            self.get_settings(snapshot),
        )

        ml_client = self.get_azureml_client()

        image = self.get_image(snapshot=snapshot)
        env = Environment(
            name=f"zenml-{snapshot.pipeline_configuration.name}",
            image=image,
        )

        entrypoint_command = (
            DynamicPipelineEntrypointConfiguration.get_entrypoint_command()
        )
        entrypoint_args = (
            DynamicPipelineEntrypointConfiguration.get_entrypoint_arguments(
                snapshot_id=snapshot.id,
                run_id=placeholder_run.id if placeholder_run else None,
            )
        )

        compute_target = create_or_get_compute(
            ml_client, settings, default_compute_name=f"zenml_{self.id}"
        )

        run_name = get_orchestrator_run_name(
            pipeline_name=snapshot.pipeline_configuration.name
        )

        command_job = command(
            name=run_name,
            command=" ".join(entrypoint_command + entrypoint_args),
            environment=env,
            environment_variables=environment,
            compute=compute_target,
            experiment_name=snapshot.pipeline_configuration.name,
            shm_size=settings.shm_size,
        )

        job = ml_client.jobs.create_or_update(command_job)

        _wait_for_completion = None
        if settings.synchronous:

            def _wait_for_completion() -> None:
                logger.info("Waiting for AzureML job to finish...")
                ml_client.jobs.stream(job.name)

        return SubmissionResult(
            metadata=self.compute_metadata(job),
            wait_for_completion=_wait_for_completion,
        )

    def submit_isolated_step(
        self, step_run_info: "StepRunInfo", environment: Dict[str, str]
    ) -> None:
        """Submits an isolated step to AzureML.

        Args:
            step_run_info: The step run information.
            environment: The environment variables to set in the execution
                environment.
        """
        logger.info(
            "Launching job for step `%s`.",
            step_run_info.pipeline_step_name,
        )

        settings = cast(
            AzureMLOrchestratorSettings,
            self.get_settings(step_run_info),
        )

        ml_client = self.get_azureml_client()

        image = step_run_info.get_image(key=ORCHESTRATOR_DOCKER_IMAGE_KEY)
        env = Environment(name=f"zenml-{step_run_info.run_name}", image=image)

        entrypoint_command = (
            StepOperatorEntrypointConfiguration.get_entrypoint_command()
        )
        entrypoint_args = (
            StepOperatorEntrypointConfiguration.get_entrypoint_arguments(
                step_name=step_run_info.pipeline_step_name,
                snapshot_id=(step_run_info.snapshot.id),
                step_run_id=str(step_run_info.step_run_id),
            )
        )

        compute_target = create_or_get_compute(
            ml_client, settings, default_compute_name=f"zenml_{self.id}"
        )

        job_name_prefix = str(step_run_info.step_run_id)[:8]
        job_name = f"{job_name_prefix}-{step_run_info.run_name}-{step_run_info.pipeline_step_name}"

        command_job = command(
            name=job_name,
            display_name=job_name,
            command=" ".join(entrypoint_command + entrypoint_args),
            environment=env,
            environment_variables=environment,
            compute=compute_target,
            experiment_name=step_run_info.pipeline.name,
            shm_size=settings.shm_size,
        )

        job = ml_client.jobs.create_or_update(command_job)
        publish_step_run_metadata(
            step_run_info.step_run_id,
            {
                self.id: {STEP_JOB_NAME_METADATA_KEY: job.name},
            },
        )
        # Include the metadata in the step run response to avoid an extra
        # API call
        step_run_info.step_run.run_metadata[STEP_JOB_NAME_METADATA_KEY] = (
            job.name
        )

    def get_isolated_step_status(
        self, step_run: "StepRunResponse"
    ) -> ExecutionStatus:
        """Gets the status of an isolated step on Vertex.

        Args:
            step_run: The step run.

        Returns:
            The status of the step run.
        """
        client = self.get_azureml_client()
        job_name = step_run.run_metadata[STEP_JOB_NAME_METADATA_KEY]
        job = client.jobs.get(job_name)
        return convert_job_status(job.status)

    def stop_isolated_step(self, step_run: "StepRunResponse") -> None:
        """Stops an isolated step on Vertex.

        Args:
            step_run: The step run.
        """
        client = self.get_azureml_client()
        job_name = step_run.run_metadata[STEP_JOB_NAME_METADATA_KEY]
        client.jobs.begin_cancel(job_name)

    def get_pipeline_run_metadata(
        self, run_id: UUID
    ) -> Dict[str, "MetadataType"]:
        """Get general component-specific metadata for a pipeline run.

        Args:
            run_id: The ID of the pipeline run.

        Returns:
            A dictionary of metadata.
        """
        try:
            ml_client = self.get_azureml_client()
            azureml_run_id = self.get_orchestrator_run_id()
            azureml_job = ml_client.jobs.get(azureml_run_id)

            metadata: Dict[str, MetadataType] = {
                METADATA_ORCHESTRATOR_RUN_ID: azureml_run_id,
            }

            url = getattr(azureml_job, "studio_url", None) or getattr(
                getattr(azureml_job, "services", {}).get("Studio", None),
                "endpoint",
                None,
            )
            if isinstance(url, str) and url:
                metadata[METADATA_ORCHESTRATOR_URL] = Uri(url)

            return metadata
        except Exception as e:
            logger.warning(
                f"Failed to fetch the Studio URL of the AzureML pipeline "
                f"job: {e}"
            )
            return {}

    def fetch_status(
        self, run: "PipelineRunResponse", include_steps: bool = False
    ) -> Tuple[
        Optional[ExecutionStatus], Optional[Dict[str, ExecutionStatus]]
    ]:
        """Refreshes the status of a specific pipeline run.

        Args:
            run: The run that was executed by this orchestrator.
            include_steps: Whether to fetch steps (not supported for AzureML).

        Returns:
            A tuple of (pipeline_status, step_statuses_dict).
            Step statuses are not supported for AzureML, so step_statuses_dict will always be None.

        Raises:
            AssertionError: If the run was not executed by to this orchestrator.
            ValueError: If it fetches an unknown state or if we can not fetch
                the orchestrator run ID.
        """
        # Make sure that the stack exists and is accessible
        if run.stack is None:
            raise ValueError(
                "The stack that the run was executed on is not available "
                "anymore."
            )

        # Make sure that the run belongs to this orchestrator
        assert (
            self.id
            == run.stack.components[StackComponentType.ORCHESTRATOR][0].id
        )

        ml_client = self.get_azureml_client()

        # Fetch the status of the PipelineJob
        if METADATA_ORCHESTRATOR_RUN_ID in run.run_metadata:
            run_id = run.run_metadata[METADATA_ORCHESTRATOR_RUN_ID]
        elif run.orchestrator_run_id is not None:
            run_id = run.orchestrator_run_id
        else:
            raise ValueError(
                "Can not find the orchestrator run ID, thus can not fetch "
                "the status."
            )
        status = ml_client.jobs.get(run_id).status

        pipeline_status = convert_job_status(status)

        # AzureML doesn't support step-level status fetching yet
        return pipeline_status, None

    def compute_metadata(self, job: Any) -> Dict[str, MetadataType]:
        """Generate run metadata based on the generated AzureML PipelineJob.

        Args:
            job: The corresponding PipelineJob object.

        Returns:
            A dictionary of metadata related to the pipeline run.
        """
        # Metadata
        metadata: Dict[str, MetadataType] = {}

        # Orchestrator Run ID
        if run_id := self._compute_orchestrator_run_id(job):
            metadata[METADATA_ORCHESTRATOR_RUN_ID] = run_id

        # URL to the AzureML's pipeline view
        if orchestrator_url := self._compute_orchestrator_url(job):
            metadata[METADATA_ORCHESTRATOR_URL] = Uri(orchestrator_url)

        return metadata

    @staticmethod
    def _compute_orchestrator_url(job: Any) -> Optional[str]:
        """Generate the Orchestrator Dashboard URL upon pipeline execution.

        Args:
            job: The corresponding PipelineJob object.

        Returns:
             the URL to the dashboard view in AzureML.
        """
        try:
            if job.studio_url:
                return str(job.studio_url)

            return None

        except Exception as e:
            logger.warning(
                f"There was an issue while extracting the pipeline url: {e}"
            )
            return None

    @staticmethod
    def _compute_orchestrator_run_id(job: Any) -> Optional[str]:
        """Generate the Orchestrator Dashboard URL upon pipeline execution.

        Args:
            job: The corresponding PipelineJob object.

        Returns:
             the URL to the dashboard view in AzureML.
        """
        try:
            if job.name:
                return str(job.name)

            return None

        except Exception as e:
            logger.warning(
                f"There was an issue while extracting the pipeline run ID: {e}"
            )
            return None



"""Add stack/component env and secrets [a71cca37e42d].

Revision ID: a71cca37e42d
Revises: 0.85.0
Create Date: 2025-05-08 09:50:39.261042

"""

import sqlalchemy as sa
import sqlmodel
from alembic import op

# revision identifiers, used by Alembic.
revision = "a71cca37e42d"
down_revision = "0.85.0"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "secret_resource",
        sa.Column("id", sqlmodel.sql.sqltypes.GUID(), nullable=False),
        sa.Column("created", sa.DateTime(), nullable=False),
        sa.Column("updated", sa.DateTime(), nullable=False),
        sa.Column("secret_id", sqlmodel.sql.sqltypes.GUID(), nullable=False),
        sa.Column("resource_id", sqlmodel.sql.sqltypes.GUID(), nullable=False),
        sa.Column("resource_type", sa.VARCHAR(length=255), nullable=False),
        sa.ForeignKeyConstraint(
            ["secret_id"],
            ["secret.id"],
            name="fk_secret_resource_secret_id_secret",
            ondelete="CASCADE",
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    with op.batch_alter_table("secret_resource", schema=None) as batch_op:
        batch_op.create_index(
            "ix_secret_resource_resource_id_resource_type_secret_id",
            ["resource_id", "resource_type", "secret_id"],
            unique=True,
        )

    with op.batch_alter_table("stack", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column("environment", sa.LargeBinary(), nullable=True)
        )

    with op.batch_alter_table("stack_component", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column("environment", sa.LargeBinary(), nullable=True)
        )

    # ### end Alembic commands ###


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("stack_component", schema=None) as batch_op:
        batch_op.drop_column("environment")

    with op.batch_alter_table("stack", schema=None) as batch_op:
        batch_op.drop_column("environment")

    with op.batch_alter_table("secret_resource", schema=None) as batch_op:
        batch_op.drop_index(
            "ix_secret_resource_resource_id_resource_type_secret_id"
        )

    op.drop_table("secret_resource")
    # ### end Alembic commands ###



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Endpoint definitions for stacks."""

from typing import List, Optional, Union
from uuid import UUID

from fastapi import APIRouter, Depends, Security
from pydantic import BaseModel

from zenml.constants import API, STACKS, VERSION_1
from zenml.models import (
    Page,
    StackFilter,
    StackRequest,
    StackResponse,
    StackUpdate,
)
from zenml.zen_server.auth import AuthContext, authorize
from zenml.zen_server.exceptions import error_response
from zenml.zen_server.rbac.endpoint_utils import (
    verify_permissions_and_delete_entity,
    verify_permissions_and_get_entity,
    verify_permissions_and_list_entities,
    verify_permissions_and_update_entity,
)
from zenml.zen_server.rbac.models import Action, ResourceType
from zenml.zen_server.rbac.utils import (
    batch_verify_permissions_for_models,
    verify_permission,
    verify_permission_for_model,
)
from zenml.zen_server.routers.projects_endpoints import workspace_router
from zenml.zen_server.utils import (
    async_fastapi_endpoint_wrapper,
    make_dependable,
    zen_store,
)

router = APIRouter(
    prefix=API + VERSION_1 + STACKS,
    tags=["stacks"],
    responses={401: error_response, 403: error_response},
)


@router.post(
    "",
    responses={401: error_response, 409: error_response, 422: error_response},
)
# TODO: the workspace scoped endpoint is only kept for dashboard compatibility
# and can be removed after the migration
@workspace_router.post(
    "/{project_name_or_id}" + STACKS,
    responses={401: error_response, 409: error_response, 422: error_response},
    deprecated=True,
    tags=["stacks"],
)
@async_fastapi_endpoint_wrapper
def create_stack(
    stack: StackRequest,
    project_name_or_id: Optional[Union[str, UUID]] = None,
    auth_context: AuthContext = Security(authorize),
) -> StackResponse:
    """Creates a stack.

    Args:
        stack: Stack to register.
        project_name_or_id: Optional name or ID of the project.
        auth_context: Authentication context.

    Returns:
        The created stack.
    """
    rbac_read_checks: List[BaseModel] = []

    # Check the service connector creation
    is_connector_create_needed = False
    for connector_id_or_info in stack.service_connectors:
        if isinstance(connector_id_or_info, UUID):
            service_connector = zen_store().get_service_connector(
                connector_id_or_info, hydrate=False
            )
            rbac_read_checks.append(service_connector)
        else:
            is_connector_create_needed = True

    # Check the component creation
    if is_connector_create_needed:
        verify_permission(
            resource_type=ResourceType.SERVICE_CONNECTOR, action=Action.CREATE
        )
    is_component_create_needed = False
    for components in stack.components.values():
        for component_id_or_info in components:
            if isinstance(component_id_or_info, UUID):
                component = zen_store().get_stack_component(
                    component_id_or_info, hydrate=False
                )
                rbac_read_checks.append(component)
            else:
                is_component_create_needed = True
    if is_component_create_needed:
        verify_permission(
            resource_type=ResourceType.STACK_COMPONENT,
            action=Action.CREATE,
        )

    if stack.secrets:
        secrets = [
            zen_store().get_secret_by_name_or_id(secret)
            for secret in stack.secrets
        ]
        rbac_read_checks.extend(secrets)
        stack.secrets = [secret.id for secret in secrets]

    batch_verify_permissions_for_models(rbac_read_checks, action=Action.READ)

    # Check the stack creation
    verify_permission_for_model(model=stack, action=Action.CREATE)

    return zen_store().create_stack(stack)


@router.get(
    "",
    responses={401: error_response, 404: error_response, 422: error_response},
)
# TODO: the workspace scoped endpoint is only kept for dashboard compatibility
# and can be removed after the migration
@workspace_router.get(
    "/{project_name_or_id}" + STACKS,
    responses={401: error_response, 404: error_response, 422: error_response},
    deprecated=True,
    tags=["stacks"],
)
@async_fastapi_endpoint_wrapper
def list_stacks(
    project_name_or_id: Optional[Union[str, UUID]] = None,
    stack_filter_model: StackFilter = Depends(make_dependable(StackFilter)),
    hydrate: bool = False,
    _: AuthContext = Security(authorize),
) -> Page[StackResponse]:
    """Returns all stacks.

    Args:
        project_name_or_id: Optional name or ID of the project to filter by.
        stack_filter_model: Filter model used for pagination, sorting,
            filtering.
        hydrate: Flag deciding whether to hydrate the output model(s)
            by including metadata fields in the response.

    Returns:
        All stacks matching the filter criteria.
    """
    return verify_permissions_and_list_entities(
        filter_model=stack_filter_model,
        resource_type=ResourceType.STACK,
        list_method=zen_store().list_stacks,
        hydrate=hydrate,
    )


@router.get(
    "/{stack_id}",
    responses={401: error_response, 404: error_response, 422: error_response},
)
@async_fastapi_endpoint_wrapper
def get_stack(
    stack_id: UUID,
    hydrate: bool = True,
    _: AuthContext = Security(authorize),
) -> StackResponse:
    """Returns the requested stack.

    Args:
        stack_id: ID of the stack.
        hydrate: Flag deciding whether to hydrate the output model(s)
            by including metadata fields in the response.

    Returns:
        The requested stack.
    """
    return verify_permissions_and_get_entity(
        id=stack_id, get_method=zen_store().get_stack, hydrate=hydrate
    )


@router.put(
    "/{stack_id}",
    responses={401: error_response, 404: error_response, 422: error_response},
)
@async_fastapi_endpoint_wrapper
def update_stack(
    stack_id: UUID,
    stack_update: StackUpdate,
    _: AuthContext = Security(authorize),
) -> StackResponse:
    """Updates a stack.

    Args:
        stack_id: Name of the stack.
        stack_update: Stack to use for the update.

    Returns:
        The updated stack.
    """
    rbac_read_checks: List[BaseModel] = []
    if stack_update.components:
        rbac_read_checks.extend(
            [
                zen_store().get_stack_component(id)
                for ids in stack_update.components.values()
                for id in ids
            ]
        )

    if stack_update.add_secrets:
        secrets = [
            zen_store().get_secret_by_name_or_id(secret)
            for secret in stack_update.add_secrets
        ]
        rbac_read_checks.extend(secrets)
        stack_update.add_secrets = [secret.id for secret in secrets]

    batch_verify_permissions_for_models(rbac_read_checks, action=Action.READ)

    return verify_permissions_and_update_entity(
        id=stack_id,
        update_model=stack_update,
        get_method=zen_store().get_stack,
        update_method=zen_store().update_stack,
    )


@router.delete(
    "/{stack_id}",
    responses={401: error_response, 404: error_response, 422: error_response},
)
@async_fastapi_endpoint_wrapper
def delete_stack(
    stack_id: UUID,
    _: AuthContext = Security(authorize),
) -> None:
    """Deletes a stack.

    Args:
        stack_id: Name of the stack.
    """
    verify_permissions_and_delete_entity(
        id=stack_id,
        get_method=zen_store().get_stack,
        delete_method=zen_store().delete_stack,
    )



#  Copyright (c) ZenML GmbH 2026. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Models representing pipeline run wait conditions."""

from datetime import datetime
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Type, Union
from uuid import UUID

from pydantic import Field

from zenml.enums import (
    RunWaitConditionLeaseMode,
    RunWaitConditionResolution,
    RunWaitConditionStatus,
    RunWaitConditionType,
)
from zenml.metadata.metadata_types import MetadataType
from zenml.models import RunMetadataFilterMixin
from zenml.models.v2.base.base import (
    BaseUpdate,
)
from zenml.models.v2.base.scoped import (
    ProjectScopedFilter,
    ProjectScopedRequest,
    ProjectScopedResponse,
    ProjectScopedResponseBody,
    ProjectScopedResponseMetadata,
    ProjectScopedResponseResources,
    UserScopedRequest,
)

if TYPE_CHECKING:
    from sqlalchemy.sql.elements import ColumnElement

    from zenml.models import PipelineRunResponse
    from zenml.models.v2.base.filter import AnySchema


class RunWaitConditionRequest(ProjectScopedRequest):
    """Request model for creating wait conditions."""

    run: UUID = Field(title="The pipeline run ID this condition belongs to.")
    name: str = Field(title="The name of the wait condition.")
    type: RunWaitConditionType = Field(title="The wait condition type.")
    question: Optional[str] = Field(
        default=None,
        title="Optional question shown to external actors.",
    )
    data_schema: Optional[Dict[str, Any]] = Field(
        default=None,
        title="Optional JSON Schema describing the expected output value.",
    )
    metadata: Dict[str, MetadataType] = Field(
        default={},
        title="Metadata of the wait condition.",
    )


class RunWaitConditionResolveRequest(UserScopedRequest):
    """Request model for resolving wait conditions."""

    resolution: RunWaitConditionResolution = Field(
        title="Resolution semantic for the waiting branch.",
    )
    result: Optional[Any] = Field(
        default=None,
        title="Optional resolved result value.",
    )


class RunWaitConditionLeaseUpdate(BaseUpdate):
    """Lease update model for wait conditions."""

    poller_instance_id: str
    poller_lease_expires_at: datetime
    mode: RunWaitConditionLeaseMode = RunWaitConditionLeaseMode.REFRESH


class RunWaitConditionResponseBody(ProjectScopedResponseBody):
    """Body model for wait condition responses."""

    type: RunWaitConditionType = Field(title="The wait condition type.")
    status: RunWaitConditionStatus = Field(
        title="The current condition status."
    )
    last_polled_at: Optional[datetime] = Field(
        default=None, title="Last lease/poll refresh timestamp."
    )
    poller_instance_id: Optional[str] = Field(
        default=None, title="Instance ID of the active poller."
    )
    poller_lease_expires_at: Optional[datetime] = Field(
        default=None, title="Lease expiration timestamp for poll liveness."
    )
    resolved_at: Optional[datetime] = Field(
        default=None, title="Timestamp when the condition was resolved."
    )
    resolved_by_user_id: Optional[UUID] = Field(
        default=None, title="User that resolved the condition."
    )


class RunWaitConditionResponseMetadata(ProjectScopedResponseMetadata):
    """Metadata model for wait condition responses."""

    question: Optional[str] = Field(
        default=None, title="User-facing question text."
    )
    run_metadata: Dict[str, MetadataType] = Field(
        default={}, title="Metadata of the wait condition."
    )
    data_schema: Optional[Dict[str, Any]] = Field(
        default=None,
        title="Optional JSON schema.",
    )
    resolution: Optional[RunWaitConditionResolution] = Field(
        default=None, title="Optional condition resolution."
    )
    result: Optional[Any] = Field(
        default=None,
        title="Optional resolved result value.",
    )


class RunWaitConditionResponseResources(ProjectScopedResponseResources):
    """Resource model for wait condition responses."""

    run: "PipelineRunResponse" = Field(
        title="Pipeline run associated with this wait condition."
    )


class RunWaitConditionResponse(
    ProjectScopedResponse[
        RunWaitConditionResponseBody,
        RunWaitConditionResponseMetadata,
        RunWaitConditionResponseResources,
    ]
):
    """Response model for run wait conditions."""

    name: str = Field(title="The name of the wait condition.")

    def get_hydrated_version(self) -> "RunWaitConditionResponse":
        """Get the hydrated version of this wait condition.

        Returns:
            an instance of the same entity with the metadata field attached.
        """
        from zenml.client import Client

        return Client().zen_store.get_run_wait_condition(self.id, hydrate=True)

    @property
    def type(self) -> RunWaitConditionType:
        """The `type` property.

        Returns:
            the value of the property.
        """
        return self.get_body().type

    @property
    def status(self) -> RunWaitConditionStatus:
        """The `status` property.

        Returns:
            the value of the property.
        """
        return self.get_body().status

    @property
    def last_polled_at(self) -> Optional[datetime]:
        """The `last_polled_at` property.

        Returns:
            the value of the property.
        """
        return self.get_body().last_polled_at

    @property
    def poller_instance_id(self) -> Optional[str]:
        """The `poller_instance_id` property.

        Returns:
            the value of the property.
        """
        return self.get_body().poller_instance_id

    @property
    def poller_lease_expires_at(self) -> Optional[datetime]:
        """The `poller_lease_expires_at` property.

        Returns:
            the value of the property.
        """
        return self.get_body().poller_lease_expires_at

    @property
    def resolved_at(self) -> Optional[datetime]:
        """The `resolved_at` property.

        Returns:
            the value of the property.
        """
        return self.get_body().resolved_at

    @property
    def resolved_by_user_id(self) -> Optional[UUID]:
        """The `resolved_by_user_id` property.

        Returns:
            the value of the property.
        """
        return self.get_body().resolved_by_user_id

    @property
    def question(self) -> Optional[str]:
        """The `question` property.

        Returns:
            the value of the property.
        """
        return self.get_metadata().question

    @property
    def run_metadata(self) -> Dict[str, MetadataType]:
        """The `run_metadata` property.

        Returns:
            the value of the property.
        """
        return self.get_metadata().run_metadata

    @property
    def data_schema(self) -> Optional[Dict[str, Any]]:
        """The `data_schema` property.

        Returns:
            the value of the property.
        """
        return self.get_metadata().data_schema

    @property
    def resolution(self) -> Optional[RunWaitConditionResolution]:
        """The `resolution` property.

        Returns:
            the value of the property.
        """
        return self.get_metadata().resolution

    @property
    def result(self) -> Optional[Any]:
        """The `result` property.

        Returns:
            the value of the property.
        """
        return self.get_metadata().result

    @property
    def run(self) -> "PipelineRunResponse":
        """The `run` property.

        Returns:
            the value of the property.
        """
        return self.get_resources().run


class RunWaitConditionFilter(ProjectScopedFilter, RunMetadataFilterMixin):
    """Filter model for wait conditions."""

    FILTER_EXCLUDE_FIELDS = [
        *ProjectScopedFilter.FILTER_EXCLUDE_FIELDS,
        *RunMetadataFilterMixin.FILTER_EXCLUDE_FIELDS,
        "resolved_by",
        "pipeline_run",
    ]
    CLI_EXCLUDE_FIELDS = [
        *ProjectScopedFilter.CLI_EXCLUDE_FIELDS,
        *RunMetadataFilterMixin.CLI_EXCLUDE_FIELDS,
    ]
    API_MULTI_INPUT_PARAMS = [
        *ProjectScopedFilter.API_MULTI_INPUT_PARAMS,
        *RunMetadataFilterMixin.API_MULTI_INPUT_PARAMS,
    ]

    pipeline_run: Optional[Union[UUID, str]] = Field(
        default=None,
        title="Filter by pipeline run ID/name.",
        union_mode="left_to_right",
    )
    type: Optional[str] = Field(
        default=None, title="Filter by condition type."
    )
    status: Optional[str] = Field(
        default=None, title="Filter by condition status."
    )
    name: Optional[str] = Field(
        default=None, title="Filter by wait condition name."
    )
    resolved_by: Optional[Union[UUID, str]] = Field(
        default=None,
        title="Filter by the name or ID of the user that resolved the condition.",
        union_mode="left_to_right",
    )
    resolved_at: Optional[Union[datetime, str]] = Field(
        default=None,
        title="Filter by the timestamp when the condition was resolved.",
    )
    resolution: Optional[str] = Field(
        default=None,
        title="Filter by condition resolution.",
    )

    def get_custom_filters(
        self, table: Type["AnySchema"]
    ) -> List["ColumnElement[bool]"]:
        """Get custom filters.

        Args:
            table: The query table.

        Returns:
            A list of custom filters.
        """
        custom_filters = super().get_custom_filters(table)

        from sqlmodel import and_

        from zenml.zen_stores.schemas import (
            PipelineRunSchema,
            RunWaitConditionSchema,
            UserSchema,
        )

        if self.resolved_by:
            resolved_by_filter = and_(
                RunWaitConditionSchema.resolved_by_user_id == UserSchema.id,
                self.generate_name_or_id_query_conditions(
                    value=self.resolved_by,
                    table=UserSchema,
                    additional_columns=["full_name"],
                ),
            )
            custom_filters.append(resolved_by_filter)

        if self.pipeline_run:
            pipeline_run_filter = and_(
                RunWaitConditionSchema.run_id == PipelineRunSchema.id,
                self.generate_name_or_id_query_conditions(
                    value=self.pipeline_run,
                    table=PipelineRunSchema,
                ),
            )
            custom_filters.append(pipeline_run_filter)

        return custom_filters



#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Pipeline execution utilities."""

import contextvars
from contextlib import contextmanager
from typing import TYPE_CHECKING, Any, Dict, Generator, Optional, Set, Union

from zenml.client import Client
from zenml.config.step_configurations import StepConfigurationUpdate
from zenml.enums import ExecutionStatus
from zenml.exceptions import RunMonitoringError
from zenml.logger import get_logger
from zenml.models import (
    PipelineRunResponse,
    PipelineSnapshotResponse,
)
from zenml.orchestrators.publish_utils import publish_failed_pipeline_run
from zenml.stack import Stack

if TYPE_CHECKING:
    StepConfigurationUpdateOrDict = Union[
        Dict[str, Any], StepConfigurationUpdate
    ]
    from zenml.steps import BaseStep

logger = get_logger(__name__)


_prevent_pipeline_execution = contextvars.ContextVar(
    "prevent_pipeline_execution", default=False
)


def should_prevent_pipeline_execution() -> bool:
    """Whether to prevent pipeline execution.

    Returns:
        Whether to prevent pipeline execution.
    """
    return _prevent_pipeline_execution.get()


@contextmanager
def prevent_pipeline_execution() -> Generator[None, None, None]:
    """Context manager to prevent pipeline execution.

    Yields:
        None.
    """
    token = _prevent_pipeline_execution.set(True)
    try:
        yield
    finally:
        _prevent_pipeline_execution.reset(token)


def submit_pipeline(
    snapshot: "PipelineSnapshotResponse",
    stack: "Stack",
    placeholder_run: Optional["PipelineRunResponse"] = None,
) -> None:
    """Submit a snapshot for execution.

    Args:
        snapshot: The snapshot to submit.
        stack: The stack on which to submit the snapshot.
        placeholder_run: An optional placeholder run for the snapshot.

    # noqa: DAR401
    Raises:
        BaseException: Any exception that happened while submitting or running
            (in case it happens synchronously) the pipeline.
    """
    # Prevent execution of nested pipelines which might lead to
    # unexpected behavior
    with prevent_pipeline_execution():
        try:
            stack.prepare_pipeline_submission(snapshot=snapshot)
            stack.submit_pipeline(
                snapshot=snapshot,
                placeholder_run=placeholder_run,
            )
        except RunMonitoringError as e:
            # Don't mark the run as failed if the error happened during
            # monitoring of the run.
            raise e.original_exception from None
        except BaseException as e:
            if (
                placeholder_run
                and not Client()
                .get_pipeline_run(placeholder_run.id, hydrate=False)
                .status.is_finished
            ):
                # We failed during/before the submission of the run, so we mark
                # the run as failed if it's still in an unfinished state.
                publish_failed_pipeline_run(placeholder_run.id)

            raise e


def compute_invocation_id(
    existing_invocations: Set[str],
    step: "BaseStep",
    custom_id: Optional[str] = None,
    allow_suffix: bool = True,
) -> str:
    """Compute the invocation ID.

    Args:
        existing_invocations: The existing invocation IDs.
        step: The step for which to compute the ID.
        custom_id: Custom ID to use for the invocation.
        allow_suffix: Whether a suffix can be appended to the invocation
            ID.

    Raises:
        RuntimeError: If no ID suffix is allowed and an invocation for the
            same ID already exists.
        RuntimeError: If no unique invocation ID can be found.

    Returns:
        The invocation ID.
    """
    base_id = id_ = custom_id or step.name

    if id_ not in existing_invocations:
        return id_

    if not allow_suffix:
        raise RuntimeError(f"Duplicate step ID `{id_}`")

    for index in range(2, 10000):
        id_ = f"{base_id}_{index}"
        if id_ not in existing_invocations:
            return id_

    raise RuntimeError("Unable to find step ID")


def skip_steps_and_prune_snapshot(
    snapshot: "PipelineSnapshotResponse",
    pipeline_run: "PipelineRunResponse",
) -> bool:
    """Skip steps and prune the snapshot.

    Args:
        snapshot: The snapshot to prune.
        pipeline_run: The pipeline run to skip steps for.

    Raises:
        RuntimeError: If the pipeline run is not a replayed run.
        RuntimeError: If a step has an upstream step that is not skipped.
        RuntimeError: If a step run request cannot be populated.

    Returns:
        Whether a pipeline run is still required.
    """
    from zenml.orchestrators.step_run_utils import StepRunRequestFactory

    if snapshot.is_dynamic:
        # In dynamic pipelines, the steps will be skipped at runtime.
        return True

    if not pipeline_run.original_run:
        raise RuntimeError(
            "Unable to skip steps because the pipeline run is not a "
            "replayed run."
        )

    logger.debug("Skipping steps and pruning snapshot.")

    client = Client()
    request_factory = StepRunRequestFactory(
        snapshot=snapshot,
        pipeline_run=pipeline_run,
        stack=client.active_stack,
    )

    explicitly_skipped_steps = set(pipeline_run.config.steps_to_skip)
    skipped_invocations: Set[str] = set()

    for invocation_id, step in snapshot.step_configurations.items():
        explicitly_skipped = invocation_id in explicitly_skipped_steps
        should_skip = request_factory.should_skip_step(invocation_id)

        if not should_skip:
            continue

        unskipped_upstream_steps = (
            set(step.spec.upstream_steps) - skipped_invocations
        )

        if unskipped_upstream_steps:
            if not explicitly_skipped:
                logger.debug(
                    "Not skipping successful step `%s` because upstream "
                    "steps `%s` are not skipped.",
                    invocation_id,
                    ", ".join(unskipped_upstream_steps),
                )
                continue

            raise RuntimeError(
                f"Unable to skip step `{invocation_id}` because it has "
                f"upstream steps `{', '.join(unskipped_upstream_steps)}` that "
                "are not skipped."
            )

        request = request_factory.create_request(invocation_id)
        try:
            request_factory.populate_request(request)
        except Exception as e:
            # We failed to populate the step run request. This might be due
            # to some input resolution error, or an error importing the step
            # source (there might be some missing dependencies). We do not want
            # the orchestrator to spin up an environment for this step, so we
            # fail early here.
            raise RuntimeError(
                "Failed to populate step run request for step "
                f"`{invocation_id}`: {str(e)}"
            ) from e

        if request.status != ExecutionStatus.SKIPPED:
            # This shouldn't happen, but just in case.
            raise RuntimeError(
                f"Expected step request `{invocation_id}` to have status "
                f"`{ExecutionStatus.SKIPPED}`, but got `{request.status}`."
            )

        client.zen_store.create_run_step(request)
        skipped_invocations.add(invocation_id)
        logger.info("Skipping step `%s`.", invocation_id)

    for invocation_id in skipped_invocations:
        # Remove the skipped step invocations from the snapshot so
        # the orchestrator does not try to run them
        snapshot.step_configurations.pop(invocation_id)

    for step in snapshot.step_configurations.values():
        for invocation_id in skipped_invocations:
            if invocation_id in step.spec.upstream_steps:
                step.spec.upstream_steps.remove(invocation_id)

    if len(snapshot.step_configurations) == 0:
        logger.info("All steps were skipped.")
        return False

    return True



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Server middlewares."""

import logging
import os
from asyncio import Lock
from asyncio.log import logger
from datetime import datetime, timedelta
from typing import Any, Set

from anyio import CapacityLimiter, to_thread
from fastapi import FastAPI, Request
from fastapi.responses import PlainTextResponse
from starlette.middleware.base import (
    BaseHTTPMiddleware,
    RequestResponseEndpoint,
)
from starlette.middleware.cors import CORSMiddleware
from starlette.responses import (
    JSONResponse,
    Response,
)
from starlette.types import ASGIApp

from zenml.analytics import source_context
from zenml.constants import (
    DEFAULT_ZENML_SERVER_REPORT_USER_ACTIVITY_TO_DB_SECONDS,
    ENV_ZENML_DEFAULT_ANALYTICS_SOURCE,
    HEALTH,
    READY,
)
from zenml.enums import SourceContextTypes
from zenml.utils.time_utils import utc_now
from zenml.zen_server.request_management import RequestContext
from zenml.zen_server.secure_headers import (
    secure_headers,
)
from zenml.zen_server.utils import (
    get_system_metrics_log_str,
    is_user_request,
    request_manager,
    server_config,
    zen_store,
)

# Track active requests with an atomic counter
active_requests_count = 0
active_requests_lock = Lock()


# Initialize last_user_activity
last_user_activity: datetime = utc_now()
last_user_activity_reported: datetime = last_user_activity + timedelta(
    seconds=-DEFAULT_ZENML_SERVER_REPORT_USER_ACTIVITY_TO_DB_SECONDS
)
last_user_activity_lock = Lock()
# Create a custom thread pool limiter with a limit of 1 thread for all
# user activity updates
last_user_activity_thread_limiter = CapacityLimiter(1)


class RequestBodyLimit(BaseHTTPMiddleware):
    """Limits the size of the request body."""

    def __init__(self, app: ASGIApp, max_bytes: int) -> None:
        """Limits the size of the request body.

        Args:
            app: The FastAPI app.
            max_bytes: The maximum size of the request body.
        """
        super().__init__(app)
        self.max_bytes = max_bytes

    async def dispatch(
        self, request: Request, call_next: RequestResponseEndpoint
    ) -> Response:
        """Limits the size of the request body.

        Args:
            request: The incoming request.
            call_next: The next function to be called.

        Returns:
            The response to the request.
        """
        if content_length := request.headers.get("content-length"):
            if int(content_length) > self.max_bytes:
                return Response(status_code=413)  # Request Entity Too Large

        try:
            return await call_next(request)
        except Exception:
            logger.exception("An error occurred while processing the request")
            return JSONResponse(
                status_code=500,
                content={"detail": "An unexpected error occurred."},
            )


class RestrictFileUploadsMiddleware(BaseHTTPMiddleware):
    """Restrict file uploads to certain paths."""

    def __init__(self, app: ASGIApp, allowed_paths: Set[str]):
        """Restrict file uploads to certain paths.

        Args:
            app: The FastAPI app.
            allowed_paths: The allowed paths.
        """
        super().__init__(app)
        self.allowed_paths = allowed_paths

    async def dispatch(
        self, request: Request, call_next: RequestResponseEndpoint
    ) -> Response:
        """Restrict file uploads to certain paths.

        Args:
            request: The incoming request.
            call_next: The next function to be called.

        Returns:
            The response to the request.
        """
        if request.method == "POST":
            content_type = request.headers.get("content-type", "")
            if (
                "multipart/form-data" in content_type
                and request.url.path not in self.allowed_paths
            ):
                return JSONResponse(
                    status_code=403,
                    content={
                        "detail": "File uploads are not allowed on this endpoint."
                    },
                )

        try:
            return await call_next(request)
        except Exception:
            logger.exception("An error occurred while processing the request")
            return JSONResponse(
                status_code=500,
                content={"detail": "An unexpected error occurred."},
            )


ALLOWED_FOR_FILE_UPLOAD: Set[str] = set()


async def track_last_user_activity(request: Request, call_next: Any) -> Any:
    """A middleware to track last user activity.

    This middleware checks if the incoming request is a user request and
    updates the last activity timestamp if it is.

    Args:
        request: The incoming request object.
        call_next: A function that will receive the request as a parameter and
            pass it to the corresponding path operation.

    Returns:
        The response to the request.
    """
    global last_user_activity
    global last_user_activity_reported
    global last_user_activity_lock

    now = utc_now()

    try:
        if is_user_request(request):
            report_user_activity = False
            async with last_user_activity_lock:
                last_user_activity = now
                if (
                    (now - last_user_activity_reported).total_seconds()
                    > DEFAULT_ZENML_SERVER_REPORT_USER_ACTIVITY_TO_DB_SECONDS
                ):
                    last_user_activity_reported = now
                    report_user_activity = True

            if report_user_activity:
                # We don't want to make a DB call here because we're in the
                # context of the asyncio event loop and it would block the
                # entire application for who knows how long.
                # We use the threadpool for it.

                request_context = request_manager().current_request

                def update_last_user_activity_timestamp() -> None:
                    logger.debug(
                        f"[{request_context.log_request_id}] API STATS - "
                        f"{request_context.log_request} "
                        f"UPDATING LAST USER ACTIVITY "
                        f"{get_system_metrics_log_str(request_context.request)}"
                    )

                    try:
                        zen_store()._update_last_user_activity_timestamp(
                            last_user_activity=last_user_activity,
                        )
                    finally:
                        logger.debug(
                            f"[{request_context.log_request_id}] API STATS - "
                            f"{request_context.log_request} "
                            f"UPDATED LAST USER ACTIVITY "
                            f"{get_system_metrics_log_str(request_context.request)}"
                        )

                await to_thread.run_sync(
                    update_last_user_activity_timestamp,
                    limiter=last_user_activity_thread_limiter,
                )

    except Exception as e:
        logger.debug(
            f"An unexpected error occurred while checking user activity: {e}"
        )

    try:
        return await call_next(request)
    except Exception:
        logger.exception("An error occurred while processing the request")
        return JSONResponse(
            status_code=500,
            content={"detail": "An unexpected error occurred."},
        )


async def infer_source_context(request: Request, call_next: Any) -> Any:
    """A middleware to track the source of an event.

    It extracts the source context from the header of incoming requests
    and applies it to the ZenML source context on the API side. This way, the
    outgoing analytics request can append it as an additional field.

    Args:
        request: the incoming request object.
        call_next: a function that will receive the request as a parameter and
            pass it to the corresponding path operation.

    Returns:
        the response to the request.
    """
    try:
        default_source = os.environ.get(
            ENV_ZENML_DEFAULT_ANALYTICS_SOURCE,
            SourceContextTypes.API.value,
        )
        s = request.headers.get(
            source_context.name,
            default=default_source,
        )
        source_context.set(SourceContextTypes(s))
    except Exception as e:
        logger.warning(
            f"An unexpected error occurred while getting the source "
            f"context: {e}"
        )
        source_context.set(SourceContextTypes.API)

    try:
        return await call_next(request)
    except Exception:
        logger.exception("An error occurred while processing the request")
        return JSONResponse(
            status_code=500,
            content={"detail": "An unexpected error occurred."},
        )


async def set_secure_headers(request: Request, call_next: Any) -> Any:
    """Middleware to set secure headers.

    Args:
        request: The incoming request.
        call_next: The next function to be called.

    Returns:
        The response with secure headers set.
    """
    try:
        response = await call_next(request)
    except Exception:
        logger.exception("An error occurred while processing the request")
        response = JSONResponse(
            status_code=500,
            content={"detail": "An unexpected error occurred."},
        )

    # If the request is for the openAPI docs, don't set secure headers
    if request.url.path.startswith("/docs") or request.url.path.startswith(
        "/redoc"
    ):
        return response

    await secure_headers().set_headers_async(response)
    return response


async def log_requests(request: Request, call_next: Any) -> Any:
    """Log requests to the ZenML server.

    Args:
        request: The incoming request object.
        call_next: A function that will receive the request as a parameter and
            pass it to the corresponding path operation.

    Returns:
        The response to the request.
    """
    global active_requests_count

    if not logger.isEnabledFor(logging.DEBUG):
        return await call_next(request)

    async with active_requests_lock:
        active_requests_count += 1

    request_context = request_manager().current_request

    logger.debug(
        f"[{request_context.log_request_id}] API STATS - "
        f"{request_context.log_request} "
        f"RECEIVED {get_system_metrics_log_str(request)}"
    )

    try:
        response = await call_next(request)

        logger.debug(
            f"[{request_context.log_request_id}] API STATS - "
            f"{response.status_code} {request_context.log_request} "
            f"took {request_context.log_duration} "
            f"{get_system_metrics_log_str(request)}"
        )

        return response
    finally:
        async with active_requests_lock:
            active_requests_count -= 1


async def record_requests(request: Request, call_next: Any) -> Any:
    """Record requests to the ZenML server.

    Args:
        request: The incoming request object.
        call_next: A function that will receive the request as a parameter and
            pass it to the corresponding path operation.

    Returns:
        The response to the request.
    """
    # Keep track of the request context in a context variable
    request_context = RequestContext(request=request)
    request_manager().current_request = request_context

    try:
        response = await call_next(request)
    except Exception:
        logger.exception("An error occurred while processing the request")
        response = JSONResponse(
            status_code=500,
            content={"detail": "An unexpected error occurred."},
        )

    return response


async def skip_health_middleware(request: Request, call_next: Any) -> Any:
    """Skip health and ready endpoints.

    Args:
        request: The incoming request.
        call_next: The next function to be called.

    Returns:
        The response to the request.
    """
    if request.url.path in [HEALTH, READY]:
        # Skip expensive processing
        return PlainTextResponse("ok")

    return await call_next(request)


def add_middlewares(app: FastAPI) -> None:
    """Add middlewares to the FastAPI app.

    Args:
        app: The FastAPI app.
    """
    app.add_middleware(BaseHTTPMiddleware, dispatch=track_last_user_activity)
    app.add_middleware(BaseHTTPMiddleware, dispatch=infer_source_context)

    app.add_middleware(
        CORSMiddleware,
        allow_origins=server_config().cors_allow_origins or [],
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )

    app.add_middleware(
        RequestBodyLimit,
        max_bytes=server_config().max_request_body_size_in_bytes,
    )
    app.add_middleware(
        RestrictFileUploadsMiddleware, allowed_paths=ALLOWED_FOR_FILE_UPLOAD
    )

    app.add_middleware(BaseHTTPMiddleware, dispatch=set_secure_headers)
    app.add_middleware(BaseHTTPMiddleware, dispatch=log_requests)
    app.add_middleware(BaseHTTPMiddleware, dispatch=record_requests)
    app.add_middleware(BaseHTTPMiddleware, dispatch=skip_health_middleware)



#  Copyright (c) ZenML GmbH 2026. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Models for the pub/sub layer."""

from datetime import datetime
from typing import Any, Literal
from uuid import UUID

from pydantic import BaseModel

from zenml.utils.enum_utils import StrEnum


class SnapshotExecutionPayload(BaseModel):
    """Payload class for snapshot async execution."""

    snapshot_id: UUID
    trigger_type: Literal["REST", "SCHEDULE", "WEBHOOK"]
    trigger_id: UUID | None = None
    scheduled_at: datetime | None = None


class CriticalEventType(StrEnum):
    """Critical event types enum."""

    READ_FAILED = "read_operation_failed"
    ACK_FAILED = "ack_operation_failed"
    ENCODE_FAILED = "encode_operation_failed"
    SEND_FAILED = "send_operation_failed"
    EXECUTE_FAILED = "execute_operation_failed"


class CriticalEvent(BaseModel, arbitrary_types_allowed=True):
    """Class capturing a critical event basic properties."""

    value: CriticalEventType
    description: str
    created_at: datetime
    exception: Exception | None = None


class MessagePayload(BaseModel):
    """Class capturing a message payload basic properties."""

    id: str
    body: SnapshotExecutionPayload | None


class MessageEnvelope(BaseModel):
    """Envelope class for consumer operations (raw payload & de-serialized payload)."""

    payload: MessagePayload
    raw_message: Any

    @property
    def id(self) -> str:
        """Implements the 'id' property.

        Returns:
            The ID of the message payload.
        """
        return self.payload.id



#  Copyright (c) ZenML GmbH 2024. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""ZenML login credentials store support."""

import os
from datetime import timedelta
from typing import Dict, List, Optional, Tuple, Union, cast

from zenml.config.global_config import GlobalConfiguration
from zenml.constants import (
    ENV_ZENML_DISABLE_CREDENTIALS_DISK_CACHING,
    handle_bool_env_var,
)
from zenml.io import fileio
from zenml.logger import get_logger
from zenml.login.credentials import APIToken, ServerCredentials, ServerType
from zenml.login.pro.constants import ZENML_PRO_API_URL
from zenml.login.pro.workspace.models import WorkspaceRead
from zenml.models import OAuthTokenResponse, ServerModel
from zenml.utils import yaml_utils
from zenml.utils.singleton import SingletonMetaClass
from zenml.utils.time_utils import utc_now_tz_aware

logger = get_logger(__name__)

CREDENTIALS_STORE_FILENAME = "credentials.yaml"

# How long to keep tokens in the credentials store after their expiration date before evicting them
TOKEN_STORE_EVICTION_TIME = 60 * 60 * 24 * 7  # 1 week


class CredentialsStore(metaclass=SingletonMetaClass):
    """Login credentials store.

    This is a singleton object that maintains a cache of all API tokens and API
    keys that are configured for the ZenML servers that the client connects to
    throughout its lifetime.

    The cache is persistent and it is backed by a `credentials.yaml` YAML file
    kept in the global configuration location. The Credentials Store cache is
    populated mainly in the following ways:

        1. when the user runs `zenml login` to authenticate to a ZenML Pro
        server, the ZenML Pro API token fetched from the web login flow is
        stored in the Credentials Store.
        2. when the user runs `zenml login` to authenticate to a ZenML Pro
        server using an API key, the ZenML Pro API key is stored in the
        Credentials Store.
        3. when the user runs `zenml login` to authenticate to a regular ZenML
        server with the web login flow, the ZenML server API token fetched
        through the web login flow is stored in the Credentials Store
        4. when the user runs `zenml login` to authenticate to any ZenML server
        using an API key, the API key is stored in the Credentials Store
        5. when the REST zen store is initialized, it starts up not yet
        authenticated. Then, if/when it needs to authenticate or re-authenticate
        to the remote server, it will use whatever form of credentials it finds
        in the Credentials Store, in order of priority:

            * if it finds an API token that is not expired (e.g. authorized
            device API tokens fetched through the web login flow or short-lived
            session API tokens fetched through some other means of
            authentication), it will use that to authenticate
            * for ZenML servers that use an API key to authenticate, it will use
            that to fetch a short-lived ZenML Pro server API token that it also
            stores in the Credentials Store
            * for ZenML Pro servers:
                * if a valid ZenML Pro API token is found, it exchanges it into
                a short lived ZenML Pro server API token
                * otherwise, if a ZenML Pro API key is found, it will use that
                to fetch a short-lived ZenML Pro server API token that it also
                stores in the Credentials Store

    Alongside credentials, the Credentials Store is also used to store
    additional server information:
        * ZenML Pro workspace information populated by the `zenml login` command
        * ZenML server information populated by the REST zen store by fetching
        the server's information endpoint after authenticating

    """

    credentials: Dict[str, ServerCredentials]
    last_modified_time: Optional[float] = None

    def __init__(self) -> None:
        """Initializes the login credentials store with values loaded from the credentials YAML file.

        CredentialsStore is a singleton class: only one instance can exist.
        Calling this constructor multiple times will always yield the same
        instance.
        """
        self.credentials = {}
        self._load_credentials()

    @classmethod
    def reset_instance(
        cls, store: Optional["CredentialsStore"] = None
    ) -> None:
        """Reset the singleton instance of the CredentialsStore.

        Args:
            store: Optional instance of the CredentialsStore to set as the
                singleton instance. If None, a new instance will be created.
        """
        current_store = cls.get_instance()
        if current_store is not None and current_store is not store:
            # Delete the credentials file from disk if it exists, otherwise
            # the credentials will be reloaded from the file when the new
            # instance is created and this call will have no effect
            current_store._delete_credentials_file()

        cls._clear(store)  # type: ignore[arg-type]
        if store:
            store._save_credentials()

    @classmethod
    def get_instance(cls) -> Optional["CredentialsStore"]:
        """Get the singleton instance of the CredentialsStore.

        Returns:
            The singleton instance of the CredentialsStore.
        """
        return cast(CredentialsStore, cls._instance())

    @property
    def _credentials_file(self) -> str:
        """Path to the file where the credentials are stored.

        Returns:
            The path to the file where the credentials are stored.
        """
        config_path = GlobalConfiguration().config_directory
        return os.path.join(config_path, CREDENTIALS_STORE_FILENAME)

    def _load_credentials(self) -> None:
        """Load the credentials from the YAML file if it exists."""
        if handle_bool_env_var(ENV_ZENML_DISABLE_CREDENTIALS_DISK_CACHING):
            return

        credentials_file = self._credentials_file
        credentials_store = {}

        if fileio.exists(credentials_file):
            try:
                credentials_store = yaml_utils.read_yaml(credentials_file)
            except Exception as e:
                logger.error(
                    f"Failed to load credentials file {credentials_file}: {e}. "
                )
            self.last_modified_time = os.path.getmtime(credentials_file)

        if credentials_store is None:
            # This can happen for example if the config file is empty
            credentials_store = {}
        elif not isinstance(credentials_store, dict):
            logger.warning(
                f"The credentials file {credentials_file} is corrupted. "
                "Creating a new credentials file."
            )
            credentials_store = {}

        self.credentials = {}
        for server_url, token_data in credentials_store.items():
            try:
                self.credentials[server_url] = ServerCredentials(**token_data)
            except ValueError as e:
                logger.warning(
                    f"Failed to load credentials for {server_url}: {e}. "
                    "Ignoring this token."
                )

    def _save_credentials(self) -> None:
        """Dump the current credentials store to the YAML file."""
        if handle_bool_env_var(ENV_ZENML_DISABLE_CREDENTIALS_DISK_CACHING):
            return
        credentials_file = self._credentials_file
        credentials_store = {
            server_url: credential.model_dump(
                mode="json", exclude_none=True, exclude_unset=True
            )
            for server_url, credential in self.credentials.items()
            # Evict tokens that have expired past the eviction time
            # and have no API key or username/password to fall back on
            if credential.api_key
            or credential.username
            and credential.password is not None
            or credential.api_token
            and (
                not credential.api_token.expires_at
                or credential.api_token.expires_at
                + timedelta(seconds=TOKEN_STORE_EVICTION_TIME)
                > utc_now_tz_aware()
            )
        }
        yaml_utils.write_yaml(credentials_file, credentials_store)
        self.last_modified_time = os.path.getmtime(credentials_file)

    def _delete_credentials_file(self) -> None:
        """Delete the credentials file."""
        if handle_bool_env_var(ENV_ZENML_DISABLE_CREDENTIALS_DISK_CACHING):
            return
        credentials_file = self._credentials_file
        if fileio.exists(credentials_file):
            fileio.remove(credentials_file)
            self.last_modified_time = None

    def check_and_reload_from_file(self) -> None:
        """Check if the credentials file has been modified and reload it if necessary."""
        if handle_bool_env_var(ENV_ZENML_DISABLE_CREDENTIALS_DISK_CACHING):
            return
        if not self.last_modified_time:
            return
        credentials_file = self._credentials_file
        try:
            last_modified_time = os.path.getmtime(credentials_file)
        except FileNotFoundError:
            # The credentials file has been deleted
            self.last_modified_time = None
            return
        if last_modified_time != self.last_modified_time:
            self._load_credentials()

    def get_password(
        self, server_url: str
    ) -> Tuple[Optional[str], Optional[str]]:
        """Retrieve the username and password from the credentials store for a specific server URL.

        Args:
            server_url: The server URL for which to retrieve the username and
                password.

        Returns:
            The stored username and password if they exist, None otherwise.
        """
        self.check_and_reload_from_file()
        credential = self.credentials.get(server_url)
        if credential:
            return credential.username, credential.password
        return None, None

    def get_api_key(self, server_url: str) -> Optional[str]:
        """Retrieve an API key from the credentials store for a specific server URL.

        Args:
            server_url: The server URL for which to retrieve the API key.

        Returns:
            The stored API key if it exists, None otherwise.
        """
        self.check_and_reload_from_file()
        credential = self.credentials.get(server_url)
        if credential:
            return credential.api_key
        return None

    def get_token(
        self, server_url: str, allow_expired: bool = False
    ) -> Optional[APIToken]:
        """Retrieve a valid token from the credentials store for a specific server URL.

        Args:
            server_url: The server URL for which to retrieve the token.
            allow_expired: Whether to allow expired tokens to be returned. The
                default behavior is to return None if a token does exist but is
                expired.

        Returns:
            The stored token if it exists and is not expired, None otherwise.
        """
        self.check_and_reload_from_file()
        credential = self.credentials.get(server_url)
        if credential:
            token = credential.api_token
            if token and (not token.expired or allow_expired):
                return token
        return None

    def get_credentials(self, server_url: str) -> Optional[ServerCredentials]:
        """Retrieve the credentials for a specific server URL.

        Args:
            server_url: The server URL for which to retrieve the credentials.

        Returns:
            The stored credentials if they exist, None otherwise.
        """
        self.check_and_reload_from_file()
        return self.credentials.get(server_url)

    def has_valid_credentials(self, server_url: str) -> bool:
        """Check if valid credentials exist for a specific server URL.

        Args:
            server_url: The server URL for which to check the authentication.

        Returns:
            True if the credentials store contains credentials that can be used
            to login to the given server URL, False otherwise.
        """
        credentials = self.get_credentials(server_url)
        if credentials and (
            credentials.has_valid_token or credentials.can_refresh_token
        ):
            return True
        return False

    def get_pro_api_key(self, pro_api_url: str) -> Optional[str]:
        """Retrieve an API key from the credentials store for a ZenML Pro API server.

        Args:
            pro_api_url: The URL of the ZenML Pro API server.

        Returns:
            The stored API key if it exists, None otherwise.
        """
        credential = self.get_pro_credentials(pro_api_url)
        if credential and credential.api_key:
            return credential.api_key
        return None

    def get_pro_token(
        self, pro_api_url: str, allow_expired: bool = False
    ) -> Optional[APIToken]:
        """Retrieve a valid token from the credentials store for a ZenML Pro API server.

        Args:
            pro_api_url: The URL of the ZenML Pro API server.
            allow_expired: Whether to allow expired tokens to be returned. The
                default behavior is to return None if a token does exist but is
                expired.

        Returns:
            The stored token if it exists and is not expired, None otherwise.
        """
        credential = self.get_pro_credentials(pro_api_url)
        if credential:
            token = credential.api_token
            if token and (not token.expired or allow_expired):
                return token
        return None

    def get_pro_credentials(
        self, pro_api_url: str
    ) -> Optional[ServerCredentials]:
        """Retrieve valid credentials from the credentials store for a ZenML Pro API server.

        Args:
            pro_api_url: The URL of the ZenML Pro API server.

        Returns:
            The stored credentials if they exist, None otherwise.
        """
        credential = self.get_credentials(pro_api_url)
        if credential and credential.type == ServerType.PRO_API:
            return credential
        return None

    def has_valid_pro_credentials(self, pro_api_url: str) -> bool:
        """Check if valid credentials exist for a ZenML Pro API server.

        Args:
            pro_api_url: The URL of the ZenML Pro API server.

        Returns:
            True if the credentials store contains credentials that can be used
            to login to the given ZenML Pro API server, False otherwise.
        """
        credentials = self.get_pro_credentials(pro_api_url)
        if credentials and (
            credentials.has_valid_token or credentials.can_refresh_token
        ):
            return True
        return False

    def clear_pro_credentials(self, pro_api_url: str) -> None:
        """Delete the token from the store for a ZenML Pro API server.

        Args:
            pro_api_url: The URL of the ZenML Pro API server.
        """
        self.clear_token(pro_api_url)

    def clear_all_credentials(self) -> None:
        """Delete all stored credentials."""
        self.credentials = {}
        self._save_credentials()

    def can_login(self, server_url: str) -> bool:
        """Check if credentials to login to the given server exist.

        Args:
            server_url: The server URL for which to check the authentication.

        Returns:
            True if the credentials store contains credentials that can be used
            to login to the given server URL, False otherwise.
        """
        credentials = self.get_credentials(server_url)
        if not credentials:
            return False

        if credentials.can_refresh_token:
            return True

        if credentials.type == ServerType.PRO:
            pro_api_url = credentials.pro_api_url or ZENML_PRO_API_URL
            pro_credentials = self.get_pro_credentials(pro_api_url)
            if pro_credentials and (
                pro_credentials.has_valid_token
                or pro_credentials.can_refresh_token
            ):
                return True

        return False

    def set_api_key(
        self,
        server_url: str,
        api_key: str,
        is_zenml_pro: bool = False,
    ) -> None:
        """Store an API key in the credentials store for a specific server URL.

        If an API token or a password is already stored for the server URL, they
        will be replaced by the API key.

        Args:
            server_url: The server URL for which the token is to be stored.
            api_key: The API key to store.
            is_zenml_pro: Whether the API key is for the ZenML Pro API.
        """
        self.check_and_reload_from_file()
        credential = self.credentials.get(server_url)
        if credential and credential.api_key != api_key:
            # Reset the API token if a new or updated API key is set, because
            # the current token might have been issued for a different account
            credential.api_token = None
            credential.api_key = api_key
            credential.username = None
            credential.password = None
            if is_zenml_pro:
                credential.pro_api_url = server_url
        else:
            self.credentials[server_url] = ServerCredentials(
                url=server_url, api_key=api_key, pro_api_url=server_url
            )

        self._save_credentials()

    def set_password(
        self,
        server_url: str,
        username: str,
        password: str,
    ) -> None:
        """Store a username and password in the credentials store for a specific server URL.

        If an API token is already stored for the server URL, it will be
        replaced by the username and password.

        Args:
            server_url: The server URL for which the token is to be stored.
            username: The username to store.
            password: The password to store.
        """
        self.check_and_reload_from_file()
        credential = self.credentials.get(server_url)
        if credential and (
            credential.username != username or credential.password != password
        ):
            # Reset the API token if a new or updated password is set, because
            # the current token might have been issued for a different account
            credential.api_token = None
            credential.username = username
            credential.password = password
            credential.api_key = None
        else:
            self.credentials[server_url] = ServerCredentials(
                url=server_url, username=username, password=password
            )

        self._save_credentials()

    def set_token(
        self,
        server_url: str,
        token_response: OAuthTokenResponse,
        is_zenml_pro: bool = False,
    ) -> APIToken:
        """Store an API token received from an OAuth2 server.

        Args:
            server_url: The server URL for which the token is to be stored.
            token_response: Token response received from an OAuth2 server.
            is_zenml_pro: Whether the token is for a ZenML Pro server.

        Returns:
            APIToken: The stored token.
        """
        self.check_and_reload_from_file()
        if token_response.expires_in:
            expires_at = utc_now_tz_aware() + timedelta(
                seconds=token_response.expires_in
            )
            # Best practice to calculate the leeway depending on the token
            # expiration time:
            #
            # - for short-lived tokens (less than 1 hour), use a fixed leeway of
            # a few seconds (e.g., 30 seconds)
            # - for longer-lived tokens (e.g., 1 hour or more), use a
            # percentage-based leeway of 5-10%
            if token_response.expires_in < 3600:
                leeway = 30
            else:
                leeway = token_response.expires_in // 20
        else:
            expires_at = None
            leeway = None

        api_token = APIToken(
            access_token=token_response.access_token,
            expires_in=token_response.expires_in,
            expires_at=expires_at,
            leeway=leeway,
            device_id=token_response.device_id,
            device_metadata=token_response.device_metadata,
        )

        credential = self.credentials.get(server_url)
        if credential:
            credential.api_token = api_token
        else:
            credential = self.credentials[server_url] = ServerCredentials(
                url=server_url, api_token=api_token
            )

        if is_zenml_pro:
            # This is how we encode that the token is for a ZenML Pro server
            credential.pro_api_url = server_url

        self._save_credentials()

        return api_token

    def set_bare_token(
        self,
        server_url: str,
        token: str,
    ) -> APIToken:
        """Store a bare API token.

        Args:
            server_url: The server URL for which the token is to be stored.
            token: The token to store.

        Returns:
            APIToken: The stored token.
        """
        self.check_and_reload_from_file()

        api_token = APIToken(
            access_token=token,
        )

        credential = self.credentials.get(server_url)
        if credential:
            credential.api_token = api_token
        else:
            self.credentials[server_url] = ServerCredentials(
                url=server_url, api_token=api_token
            )

        self._save_credentials()

        return api_token

    def update_server_info(
        self,
        server_url: str,
        server_info: Union[ServerModel, WorkspaceRead],
    ) -> None:
        """Update the server information stored for a specific server URL.

        Args:
            server_url: The server URL for which the server information is to be
                updated.
            server_info: Updated server information.
        """
        self.check_and_reload_from_file()

        credential = self.credentials.get(server_url)
        if not credential:
            # No credentials stored for this server URL, nothing to update
            return

        credential.update_server_info(server_info)
        self._save_credentials()

    def clear_token(self, server_url: str) -> None:
        """Delete a token from the store for a specific server URL.

        Args:
            server_url: The server URL for which to delete the token.
        """
        self.check_and_reload_from_file()

        if server_url in self.credentials:
            credential = self.credentials[server_url]
            if (
                not credential.api_key
                and not credential.username
                and not credential.password is not None
            ):
                # Only delete the credential entry if there is no API key or
                # username/password to fall back on
                del self.credentials[server_url]
            else:
                credential.api_token = None
            self._save_credentials()

    def clear_credentials(self, server_url: str) -> None:
        """Delete all credentials from the store for a specific server URL.

        Args:
            server_url: The server URL for which to delete the credentials.
        """
        self.check_and_reload_from_file()

        if server_url in self.credentials:
            del self.credentials[server_url]
            self._save_credentials()

    def list_credentials(
        self, type: Optional[ServerType] = None
    ) -> List[ServerCredentials]:
        """Get all credentials stored in the credentials store.

        Args:
            type: Optional server type to filter the credentials by.

        Returns:
            A list of all credentials stored in the credentials store.
        """
        self.check_and_reload_from_file()

        credentials = list(self.credentials.values())

        if type is not None:
            credentials = [c for c in credentials if c and c.type == type]

        return credentials


def get_credentials_store() -> CredentialsStore:
    """Get the global credentials store instance.

    Returns:
        The global credentials store instance.
    """
    return CredentialsStore()



#  Copyright (c) ZenML GmbH 2023. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Entrypoint configuration for ZenML Databricks pipeline steps."""

import os
import sys
from importlib.metadata import distribution
from typing import Any, Dict, List

from zenml.entrypoints.step_entrypoint_configuration import (
    StepEntrypointConfiguration,
)

WHEEL_PACKAGE_OPTION = "wheel_package"
DATABRICKS_JOB_ID_OPTION = "databricks_job_id"
ENV_ZENML_DATABRICKS_ORCHESTRATOR_RUN_ID = (
    "ZENML_DATABRICKS_ORCHESTRATOR_RUN_ID"
)


class DatabricksEntrypointConfiguration(StepEntrypointConfiguration):
    """Entrypoint configuration for ZenML Databricks pipeline steps.

    The only purpose of this entrypoint configuration is to reconstruct the
    environment variables that exceed the maximum length of 256 characters
    allowed for Databricks Processor steps from their individual components.
    """

    @classmethod
    def get_entrypoint_options(cls) -> Dict[str, bool]:
        """Gets all options required for running with this configuration.

        Returns:
            The superclass options as well as an option for the wheel package.
        """
        return super().get_entrypoint_options() | {
            WHEEL_PACKAGE_OPTION: True,
            DATABRICKS_JOB_ID_OPTION: True,
        }

    @classmethod
    def get_entrypoint_arguments(
        cls,
        **kwargs: Any,
    ) -> List[str]:
        """Gets all arguments that the entrypoint command should be called with.

        The argument list should be something that
        `argparse.ArgumentParser.parse_args(...)` can handle (e.g.
        `["--some_option", "some_value"]` or `["--some_option=some_value"]`).
        It needs to provide values for all options returned by the
        `get_entrypoint_options()` method of this class.

        Args:
            **kwargs: Kwargs, must include the step name.

        Returns:
            The superclass arguments as well as arguments for the wheel package.
        """
        return super().get_entrypoint_arguments(**kwargs) + [
            f"--{WHEEL_PACKAGE_OPTION}",
            kwargs[WHEEL_PACKAGE_OPTION],
            f"--{DATABRICKS_JOB_ID_OPTION}",
            kwargs[DATABRICKS_JOB_ID_OPTION],
        ]

    def run(self) -> None:
        """Runs the step."""
        # Get the wheel package and add it to the sys path
        wheel_package = self.entrypoint_args[WHEEL_PACKAGE_OPTION]

        dist = distribution(wheel_package)
        project_root = os.path.join(
            str(dist.locate_file(".")), str(wheel_package)
        )

        if project_root not in sys.path:
            sys.path.insert(0, project_root)
            sys.path.insert(-1, project_root)

        # Get the job id and add it to the environment
        databricks_job_id = self.entrypoint_args[DATABRICKS_JOB_ID_OPTION]
        os.environ[ENV_ZENML_DATABRICKS_ORCHESTRATOR_RUN_ID] = (
            databricks_job_id
        )

        # Run the step
        super().run()



#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Entrypoint configuration to run a dynamic pipeline."""

from typing import Any, Dict, List
from uuid import UUID

from zenml.client import Client
from zenml.entrypoints.base_entrypoint_configuration import (
    BaseEntrypointConfiguration,
)
from zenml.execution.pipeline.dynamic.runner import DynamicPipelineRunner
from zenml.integrations.registry import integration_registry

RUN_ID_OPTION = "run_id"


class DynamicPipelineEntrypointConfiguration(BaseEntrypointConfiguration):
    """Entrypoint configuration to run a dynamic pipeline."""

    @classmethod
    def get_entrypoint_options(cls) -> Dict[str, bool]:
        """Gets all options required for running with this configuration.

        Returns:
            All options required for running with this configuration.
        """
        return super().get_entrypoint_options() | {RUN_ID_OPTION: False}

    @classmethod
    def get_entrypoint_arguments(
        cls,
        **kwargs: Any,
    ) -> List[str]:
        """Gets all arguments that the entrypoint command should be called with.

        Args:
            **kwargs: Keyword arguments.

        Returns:
            All arguments that the entrypoint command should be called with.
        """
        args = super().get_entrypoint_arguments(**kwargs)
        if run_id := kwargs.get(RUN_ID_OPTION, None):
            args.extend([f"--{RUN_ID_OPTION}", str(run_id)])
        return args

    def run(self) -> None:
        """Prepares the environment and runs the configured dynamic pipeline."""
        snapshot = self.snapshot

        # Activate all the integrations. This makes sure that all materializers
        # and stack component flavors are registered.
        integration_registry.activate_integrations()

        self.prepare_code_environment()

        run = None
        if run_id := self.entrypoint_args.get(RUN_ID_OPTION, None):
            run = Client().get_pipeline_run(UUID(run_id))

        runner = DynamicPipelineRunner(snapshot=snapshot, run=run)
        runner.run_pipeline()



"""Add status reason [83ef3cb746a5].

Revision ID: 83ef3cb746a5
Revises: aae4eed923b5
Create Date: 2025-09-04 15:41:30.829628

"""

import sqlalchemy as sa
import sqlmodel
from alembic import op

# revision identifiers, used by Alembic.
revision = "83ef3cb746a5"
down_revision = "aae4eed923b5"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("pipeline_run", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                "status_reason",
                sqlmodel.sql.sqltypes.AutoString(),
                nullable=True,
            )
        )

    # ### end Alembic commands ###


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("pipeline_run", schema=None) as batch_op:
        batch_op.drop_column("status_reason")

    # ### end Alembic commands ###



#  Copyright (c) ZenML GmbH 2024. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Endpoint definitions for logs."""

from uuid import UUID

from fastapi import APIRouter, Security

from zenml.constants import (
    API,
    LOGS,
    VERSION_1,
)
from zenml.exceptions import IllegalOperationError
from zenml.models import LogsRequest, LogsResponse, LogsUpdate
from zenml.zen_server.auth import AuthContext, authorize
from zenml.zen_server.exceptions import error_response
from zenml.zen_server.rbac.endpoint_utils import (
    verify_permissions_and_create_entity,
    verify_permissions_and_update_entity,
)
from zenml.zen_server.rbac.models import Action
from zenml.zen_server.rbac.utils import (
    batch_verify_permissions_for_models,
    dehydrate_response_model,
    verify_permission_for_model,
)
from zenml.zen_server.utils import (
    async_fastapi_endpoint_wrapper,
    zen_store,
)

router = APIRouter(
    prefix=API + VERSION_1 + LOGS,
    tags=["logs"],
    responses={401: error_response, 403: error_response},
)


@router.post(
    "",
    responses={401: error_response, 409: error_response, 422: error_response},
)
@async_fastapi_endpoint_wrapper
def create_logs(
    logs: LogsRequest,
    _: AuthContext = Security(authorize),
) -> LogsResponse:
    """Create a new log model.

    Args:
        logs: The log model to create.

    Returns:
        The created log model.
    """
    if logs.pipeline_run_id:
        verify_permission_for_model(
            model=zen_store().get_run(
                run_id=logs.pipeline_run_id, hydrate=False
            ),
            action=Action.UPDATE,
        )
    elif logs.step_run_id:
        step = zen_store().get_run_step(logs.step_run_id)
        verify_permission_for_model(
            model=zen_store().get_run(
                run_id=step.pipeline_run_id, hydrate=False
            ),
            action=Action.UPDATE,
        )

    read_verify_models = []
    if logs.artifact_store_id:
        read_verify_models.append(
            zen_store().get_stack_component(
                component_id=logs.artifact_store_id, hydrate=False
            )
        )
    if logs.log_store_id:
        read_verify_models.append(
            zen_store().get_stack_component(
                component_id=logs.log_store_id, hydrate=False
            )
        )

    batch_verify_permissions_for_models(
        models=read_verify_models,
        action=Action.READ,
    )

    return verify_permissions_and_create_entity(
        request_model=logs,
        create_method=zen_store().create_logs,
    )


@router.get(
    "/{logs_id}",
    responses={401: error_response, 404: error_response, 422: error_response},
)
@async_fastapi_endpoint_wrapper
def get_logs(
    logs_id: UUID,
    hydrate: bool = True,
    _: AuthContext = Security(authorize),
) -> LogsResponse:
    """Returns the requested log model.

    Args:
        logs_id: ID of the log model.
        hydrate: Flag deciding whether to hydrate the output model(s)
            by including metadata fields in the response.

    Returns:
        The requested log model.

    Raises:
        IllegalOperationError: If the logs are not associated
            with a pipeline run or step run before fetching.
    """
    logs = zen_store().get_logs(logs_id, hydrate=True)

    if logs.pipeline_run_id:
        verify_permission_for_model(
            model=zen_store().get_run(
                run_id=logs.pipeline_run_id, hydrate=False
            ),
            action=Action.READ,
        )
    elif logs.step_run_id:
        step = zen_store().get_run_step(
            step_run_id=logs.step_run_id, hydrate=False
        )
        verify_permission_for_model(
            model=zen_store().get_run(
                run_id=step.pipeline_run_id, hydrate=False
            ),
            action=Action.READ,
        )
    else:
        raise IllegalOperationError(
            "Logs must be associated with a pipeline run or step run "
            "before fetching."
        )

    if hydrate is False:
        logs.metadata = None

    return dehydrate_response_model(logs)


@router.put(
    "/{logs_id}",
    responses={401: error_response, 404: error_response, 422: error_response},
)
@async_fastapi_endpoint_wrapper
def update_logs(
    logs_id: UUID,
    logs_update: LogsUpdate,
    _: AuthContext = Security(authorize),
) -> LogsResponse:
    """Update an existing log model.

    Args:
        logs_id: ID of the log model to update.
        logs_update: Update to apply to the log model.

    Returns:
        The updated log model.
    """
    if logs_update.pipeline_run_id:
        verify_permission_for_model(
            model=zen_store().get_run(
                run_id=logs_update.pipeline_run_id, hydrate=False
            ),
            action=Action.UPDATE,
        )
    elif logs_update.step_run_id:
        step = zen_store().get_run_step(
            step_run_id=logs_update.step_run_id, hydrate=False
        )
        verify_permission_for_model(
            model=zen_store().get_run(
                run_id=step.pipeline_run_id, hydrate=False
            ),
            action=Action.UPDATE,
        )

    return verify_permissions_and_update_entity(
        id=logs_id,
        update_model=logs_update,
        get_method=zen_store().get_logs,
        update_method=zen_store().update_logs,
    )



#  Copyright (c) ZenML GmbH 2023. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Internal BaseStep subclass used by the step decorator."""

from typing import Any, Optional

from zenml.config.source import Source
from zenml.steps import BaseStep, step


def remove_decorator_from_source_code(
    source_code: str, decorator_name: str
) -> str:
    """Remove a decorator from the source code of a function.

    Args:
        source_code: The source code of the function.
        decorator_name: The name of the decorator.

    Returns:
        The source code of the function without the decorator.
    """
    import ast

    # Remove potential indentation from the source code
    source_code_lines = source_code.split("\n")
    if source_code_lines:
        first_line = source_code_lines[0]
        leading_spaces = len(first_line) - len(first_line.lstrip())
        source_code_lines = [
            line[leading_spaces:] if len(line) >= leading_spaces else line
            for line in source_code_lines
        ]

    module = ast.parse("\n".join(source_code_lines))
    if not module.body or not isinstance(module.body[0], ast.FunctionDef):
        return source_code

    function_def = module.body[0]
    if not function_def.decorator_list:
        return source_code

    for decorator in function_def.decorator_list:
        lineno = None
        end_lineno = None

        if isinstance(decorator, ast.Call) and isinstance(
            decorator.func, ast.Name
        ):
            if decorator.func.id == decorator_name:
                lineno = decorator.lineno
                end_lineno = decorator.end_lineno
        elif isinstance(decorator, ast.Name):
            if decorator.id == decorator_name:
                lineno = decorator.lineno
                end_lineno = decorator.end_lineno
        else:
            continue

        if lineno is not None and end_lineno is not None:
            # The line numbers are 1-indexed, so we need to
            # subtract 1
            source_code_lines = (
                source_code_lines[: lineno - 1]
                + source_code_lines[end_lineno:]
            )
            source_code = "\n".join(source_code_lines)
            break

    return source_code


class _DecoratedStep(BaseStep):
    """Internal BaseStep subclass used by the step decorator."""

    def _get_step_decorator_name(self) -> Optional[str]:
        """The name of the step decorator.

        Returns:
            The name of the step decorator.
        """
        decorator_names = [
            key
            for key, value in self.entrypoint.__globals__.items()
            if value is step
        ]
        if not decorator_names:
            return None

        return decorator_names[0]

    @property
    def source_code_cache_value(self) -> str:
        """The source code cache value of this step.

        Returns:
            The source code cache value of this step.
        """
        try:
            if decorator_name := self._get_step_decorator_name():
                return remove_decorator_from_source_code(
                    source_code=self.source_code,
                    decorator_name=decorator_name,
                )
        except Exception:
            pass

        return super().source_code_cache_value

    @property
    def source_object(self) -> Any:
        """The source object of this step.

        Returns:
            The source object of this step.
        """
        return self.entrypoint

    def resolve(self) -> "Source":
        """Resolves the step.

        Returns:
            The step source.
        """
        from zenml.utils import source_utils

        return source_utils.resolve(self.entrypoint, skip_validation=True)



"""Cache key index [4dd9d3afd2c0].

Revision ID: 4dd9d3afd2c0
Revises: 0.91.1
Create Date: 2025-11-13 11:43:34.498878

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = "4dd9d3afd2c0"
down_revision = "0.91.1"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("step_run", schema=None) as batch_op:
        batch_op.create_index(
            "ix_step_run_cache_key", ["cache_key"], unique=False
        )

    # ### end Alembic commands ###


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("step_run", schema=None) as batch_op:
        batch_op.drop_index("ix_step_run_cache_key")

    # ### end Alembic commands ###
